'''Searches represent predefined sets of critera allowing to perform database
   searches. A searched is always attached to a given class from the model.'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from DateTime import DateTime

from appy.px import Px
from appy.model.batch import Batch
from appy.ui.criteria import Criteria
from appy.ui.template import Template
from appy.utils import string as sutils
from appy.model.utils import Object as O
from appy.model.fields.group import Group
from appy.model.searches.modes import Mode
from appy.model.searches import initiators

# Error messages - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
WRONG_FIELD = 'Field "%s" does not exist on %s.'
REPLAY_S    = 'Class "%s": replaying search "%s"...'
REPLAY_OK   = 'Fetched %d object(s).'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class UiSearch:
    '''Instances of this class are generated on-the-fly for manipulating a
       Search instance from the User Interface.'''
    Mode = Mode

    # Rendering a search
    view = Px('''
     <div class="portletSearch">
      <a href=":'%s?className=%s&amp;search=%s' % \
                 (queryUrl, className, search.name)"
         class=":'current' if search.name == currentSearch else ''"
         onclick="clickOn(this)"
         title=":search.translatedDescr">:search.translated</a>
     </div>''')

    # Render search results
    pxResult = Px('''
     <x var="layout='view';
             batch=mode.batch;
             empty=mode.empty;
             search=uiSearch.search;
             showNewSearch=showNewSearch|True;
             showHeaders=showHeaders|True;
             specific=uiSearch.getResultsTop(mode, ajax)">

     <!-- Application, class-specific code before displaying results -->
     <x if="specific">::specific</x>

     <!-- Search results -->
     <div id=":mode.hook">
      <script>::mode.getAjaxData()</script>

      <!-- Pod templates -->
      <x if="not empty and not popup and search.showPods">
       <table var="fields=class_.getListPods()"
              if="fields and mode.objects" align=":dright">
        <tr>
         <td var="o=mode.objects[0]"
             for="field in fields" var2="fieldName=field.name"
             class=":not loop.field.last and 'pod' or ''">:field.pxRender</td>
        </tr>
       </table>
      </x>

      <!-- Title -->
      <div if="not popup and search.showTitle" class="pageTitle topSpaceS">
       <x>::uiSearch.translated</x>
       <x if="mode.batch and not empty">
        <img src=":url('arrowsA.svg')" class="iconS"
             style="transform: rotate(270deg)"/>
        <x>:mode.batch.total</x>
       </x>
       <x if="search.add and guard.mayInstantiate(class_)"
          var2="buttonType='small'; viaPopup=search.viaPopup; label=None;
                nav=mode.getNavInfo(batch.total)">:class_.pxAdd</x>
      </div>
      <table width="100%">
       <tr valign="top">
        <!-- Search description -->
        <td>
         <x if="search.showTitle">
          <div class="discreet"
               if="uiSearch.translatedDescr">:uiSearch.translatedDescr</div>
          <!-- Perform a new search -->
          <a if="showNewSearch and mode.newSearchUrl"
             href=":mode.newSearchUrl">:_('search_new')</a>
          <x if="mode.fromRef and not mode.inField">&nbsp;&mdash;&nbsp; 
           <a href=":mode.getRefUrl()">:_('goto_source')</a></x>
         </x>
        </td>
        <!-- (Top) navigation -->
        <td if="mode.batch and uiSearch.showTopNav"
            align=":search.navAlign">:mode.batch.pxNavigate</td>
       </tr>
      </table>

      <!-- Results -->
      <x if="not empty" var2="currentNumber=0">:mode.px</x>

      <!-- (Bottom) navigation -->
      <div if="mode.batch and uiSearch.showBottomNav"
           align=":search.navAlign">:mode.batch.pxNavigate</div>

      <!-- No result -->
      <x if="empty">
       <x>::_('query_no_result')</x>
       <x if="showNewSearch and mode.newSearchUrl"><br/><i class="discreet">
         <a href=":mode.newSearchUrl">:_('search_new')</a></i></x>
      </x>
    </div></x>''')

    def __init__(self, search, tool, ctx, initiator=None, name=None):
        self.search = search
        self.container = search.container
        self.tool = tool
        self.req = tool.req
        self.dir = ctx.dir
        self.popup = ctx.popup
        # "name" can be more than the p_search name, ie, if the search is
        # defined in a field.
        self.name = name or search.name
        self.type = 'search'
        self.colspan = search.colspan
        self.showActions = search.showActions
        self.actionsDisplay = search.actionsDisplay
        self.showTopNav = search.showNav in ('top', 'both')
        self.showBottomNav = search.showNav in ('bottom', 'both')
        className = self.container.name
        if search.translated:
            self.translated = search.translated
            self.translatedDescr = search.translatedDescr or ''
        else:
            # The label may be specific in some special cases
            labelDescr = ''
            if search.name == 'allSearch': label = '%s_plural' % className
            elif search.name == 'customSearch': label = 'search_results'
            elif not search.name: label = None
            else:
                label = '%s_%s' % (className, search.name)
                labelDescr = label + '_descr'
            _ = tool.translate
            self.translated = label and _(label) or ''
            self.translatedDescr = labelDescr and _(labelDescr) or ''
        # Strip the description (a single space may be present)
        self.translatedDescr = self.translatedDescr.strip()
        # An initiator instance if the search is in a popup
        self.initiator = initiator
        # When search results are shown in a popup, checkboxes must be present
        # even when not shown. Indeed, we want them in the DOM because object
        # ids are stored on it.
        if initiator:
            self.checkboxes = True
            self.checkboxesDefault = False
        else:
            cb = search.checkboxes
            cb = cb(tool) if callable(cb) else cb
            self.checkboxes = cb
            self.checkboxesDefault = search.checkboxesDefault

    def getRootHook(self):
        '''If there is an initiator, return the hook as defined by it. Else,
           return the name of the search.'''
        init = self.initiator
        return init.popupHook if init else (self.search.name or 'search')

    def showCheckboxes(self):
        '''When must checkboxes be shown ?'''
        init = self.initiator
        return init.showCheckboxes() if init else self.checkboxes

    def getCbJsInit(self, hook):
        '''Returns the code that creates JS data structures for storing the
           status of checkboxes for every result of this search.'''
        default = self.checkboxesDefault and 'unchecked' or 'checked'
        return '''var node=findNode(this, '%s');
                  node['_appy_objs_cbs'] = {};
                  node['_appy_objs_sem'] = '%s';''' % (hook, default)

    def getModes(self):
        '''Gets all the modes applicable when displaying search results (being
           instances of p_class_) via this search (p_self). r_ is a list of
           names and not a list of Mode instances.'''
        r = self.search.resultModes or self.container.getResultModes() or \
            Mode.default
        return r if not callable(r) else r(self.tool)

    def getMode(self):
        '''Gets the current mode'''
        return Mode.get(self)

    def getTitleMode(self, popup):
        '''How titles to search results, being instances of p_class_, must be
           rendered ?'''
        # Possible r_return values are the following.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # If...    | Every search result's title is rendered as ...
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # "link"   | links allowing to go to instances' view pages;
        # "dlink"  | links, as for "link", but the link is wrapped in a "div";
        # "select" | objects, that can be selected from a popup;
        # "text"   | simple, unclickable text.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        if popup: return 'select'
        # Check if the title mode is specified on the container class
        mode = self.container.getTitleMode()
        if not mode: return 'link'
        return mode if not callable(mode) else mode(self.tool)

    def getResultsTop(self, mode, ajax):
        '''If p_class_ defines something to display on the results page just
           before displaying search results, returns it.'''
        # Get this only on the main page, not when ajax-refreshing search
        # results.
        if ajax: return
        return self.container.getResultsTop(self.tool, self.search, mode)

    def highlight(self, text):
        '''Highlight search results within p_text'''
        return Criteria.highlight(self.tool.H(), text)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Search:
    '''Used for specifying a search for a given class'''

    traverse = {}
    # Special initiator in use when an object is being created from a Search. It
    # has nothing to do with popup-oriented initiator as defined hereafter.
    initiator = initiators.SearchInitiator

    # Parameters needed to replay a search (see m_replay)
    replayParams = ('className', 'search', 'sortKey', 'sortOrder', 'filters')

    def __init__(self, name=None, group=None, sortBy='title', sortOrder='asc',
                 maxPerPage=30, default=False, colspan=1, viaPopup=None,
                 show=True, showActions='all', actionsDisplay='block',
                 showPods=True, showTitle=True, showFilters=True,
                 showNav='both', navAlign='right', translated=None,
                 translatedDescr=None, checkboxes=False, checkboxesDefault=True,
                 container=None, add=False, resultModes=None, shownInfo=None,
                 actions=None, rowAlign='top', pageLayoutOnView=None, **fields):
        # "name" is mandatory, excepted in some special cases (ie, when used as
        # "select" param for a Ref field).
        self.name = name
        # Searches may be visually grouped in the portlet
        self.group = Group.get(group)
        self.sortBy = sortBy
        self.sortOrder = sortOrder
        self.maxPerPage = maxPerPage
        # If this search is the default one, it will be triggered by clicking
        # on main link.
        self.default = default
        self.colspan = colspan
        # Attribute "viaPopup" has the same semantics as its homonym attribute
        # on class appy.model.fields.Ref.
        self.viaPopup = viaPopup
        # If a translated name or description is already given here, we will
        # use it instead of trying to translate from labels.
        self.translated = translated
        self.translatedDescr = translatedDescr
        # Condition for showing or not this search
        self.show = show
        # Attributes "showActions" and "actionsDisplay" are similar to their
        # homonyms on the Ref class.
        self.showActions = showActions
        self.actionsDisplay = actionsDisplay
        # Various parts of search results may be shown or not
        self.showPods = showPods
        self.showTitle = showTitle # The title, description, results count...
        self.showFilters = showFilters # Filters, in grid mode
        self.showNav = showNav # May be 'top', 'bottom', 'both' or None
        self.navAlign = navAlign # Can be 'left', 'center' or 'right'
        # In dict "fields", keys are names of indexed fields and values are
        # simple search values or terms built with database operators.
        self.fields = fields
        # Must checkboxes be shown for every object of the search result ?
        self.checkboxes = checkboxes
        # Default value for checkboxes
        self.checkboxesDefault = checkboxesDefault
        # Most of the time, we know what is the class whose instances must be
        # searched. When it is not the case, the p_container can be explicitly
        # specified.
        if container and (container.__class__.__name__ != 'Class'):
            container = container.meta
        self.container = container
        # Is it possible to create new "container" instances from this search ?
        self.add = add
        # There can be various ways to display query results
        self.resultModes = resultModes
        # Similar to the homonym Ref attribute, "shownInfo" defines the columns
        # that must be shown on lists of result objects (mode "List" only). If
        # not specified, class's "listColumns" attributes is used.
        self.shownInfo = shownInfo
        # Specify here Action fields that must be shown as custom actions that
        # will be triggered on search results.
        self.actions = actions
        # When search resulrs are rendered as a list, this attribute specifies
        # each row's vertical alignment.
        self.rowAlign = rowAlign
        # You may override the page layout used when accessing, on the "view"
        # layout, every search result, by using the following attribute.
        self.pageLayoutOnView = pageLayoutOnView

    def init(self, class_):
        '''Lazy search initialisation'''
        self.container = class_

    def ui(self, o, ctx):
        '''Gets a UiSearch instance corresponding to this search'''
        return UiSearch(self, o, ctx)

    def isShowable(self, tool):
        '''Is this Search instance showable ?'''
        class_ = self.container
        r = self.show
        return r if not callable(r) else \
               tool.H().methods.call(tool, self.show, class_=class_)

    def getSessionKey(self):
        '''Returns the name of the key storing siblings of a given object in the
           context of this search.'''
        return '%s_%s' % (self.container.name, self.name)

    mergeAttributes = ('sortBy', 'sortOrder', 'showActions', 'actionsDisplay',
                       'actions', 'maxPerPage', 'resultModes', 'shownInfo',
                       'checkboxes', 'checkboxesDefault')

    def merge(self, other):
        '''Merge parameters from another search in p_other'''
        self.fields.update(other.fields)
        for name in self.mergeAttributes:
            setattr(self, name, getattr(other, name))

    def getActions(self, tool):
        '''Get the actions triggerable on p_self's results'''
        actions = self.actions
        if not actions: return
        r = []
        for action in actions:
            show = action.show
            show = show(tool) if callable(show) else show
            if show:
                r.append(action)
        return r

    def getField(self, name, handler):
        '''Gets the field whose name is p_name'''
        className = self.container.name
        return handler.server.model.classes.get(className).fields.get(name)

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  The "run" method
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    def run(self, handler, batch=True, start=0, ids=False, secure=True,
            sortBy=None, sortOrder='asc', filters=None, refObject=None,
            refField=None, other=None):
        '''Based on p_self's parameters, and this method's attributes, perform a
           search in the corresponding catalog ans return the results.'''
        # If p_batch is True, the result is a appy.model.batch.Batch instance
        # returning only a subset of the results starting at p_start and
        # containing at most p_self.maxPerPage results. Else, it is a list of
        # objects (or object ids if p_ids is True) representing the complete
        # result set. If p_ids is True, p_batch is ignored and implicitly
        # considered being False.
        # ~~~
        # p_secure is transmitted and documented in called method
        # appy.database.catalog.Catalog.search (called via method
        # appy.database.Database.search).
        # ~~~
        # The result is sorted according to the potential sort key and order
        # ('asc'ending or 'desc'ending) as defined in p_self.sortBy and
        # p_self.sortOrder. If p_sortBy is not None, p_sortBy and p_sortOrder
        # override p_self's homonym attributes.
        # ~~~
        # p_filters alter search parameters according to selected filters in the
        # ui.
        # ~~~
        # If p_refObject and p_refField are given, the search is limited to the
        # objects being referenced from p_refObject through p_refField.
        # ~~~
        # If p_other is not None, it is another Search instance whose parameters
        # will be merged with p_self's parameters.
        # ~~~
        # Prepare search parameters
        className = self.container.name
        params = self.fields.copy()
        # Use the search's default sort key and order if not passed as args
        if not sortBy:
            sortBy = self.sortBy
            sortOrder = self.sortOrder
        # Take p_filters into account
        if filters:
            for name, value in filters.items():
                if not value: continue
                field = self.getField(name, handler)
                value = field.getSearchValue(None, value=value)
                if value: params[name] = value
        # Manage a potential Ref field
        if refObject and refField.back:
            backName = refField.back.name
            if refField.composite:
                # Use the "cid" index (object containment)
                params['cid'] = '%d_%s' % (refObject.iid, backName)
            else:
                # We suppose the Ref field is indexed
                params[backName] = refObject.iid
        # Perform the search in the corresponding catalog, getting IIDs of
        # matched objects.
        database = handler.server.database
        r = database.search(handler, className, ids=True, secure=secure,
                            sortBy=sortBy, sortOrder=sortOrder, **params)
        # Return object IIDs if requested
        if ids: return r
        # Return the complete set of matched objects if p_batch is False
        if not batch: return database.getObjects(handler, r, className)
        # Create a Batch and populate it with the appropriate subset of objects
        batch = Batch(total=len(r), size=self.maxPerPage, start=start)
        batch.setObjects(database.getObjects(handler, r, className,
                                            start=batch.start, size=batch.size))
        return batch

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Class methods
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    @classmethod
    def get(class_, name, tool, modelClass, ctx, ui=True):
        '''Gets the Search instance (or a UiSearch instance if p_ui is True)
           corresponding to the search named p_name, on class p_modelClass.'''
        initiator = None
        req = tool.req
        if name == 'customSearch':
            # It is a custom search
            if req.source == 'searchForm':
                # Search criteria are in the request, sent from the search form
                co = Criteria(tool)
                co.getFromRequest(modelClass)
                criteria = co.criteria
                # Set them on the request object: it will be used to store them
                # in the browser's session storage (see modes.py).
                req.criteria = co.asString()
            else:
                # Search criteria are carried in the request as a marshalled
                # string in key "criteria".
                criteria = Criteria.readFromRequest(tool.H()) or {}
            # Create the Search instance, ommitting the "_ref" attribute, that
            # does not correspond to any field.
            if '_ref' in criteria: del(criteria['_ref'])
            r = Search('customSearch', container=modelClass, **criteria)
            # Avoid name clash
            if 'group' in criteria: r.fields['group'] = criteria['group']
            # Take into account default params for the advanced search
            advanced = modelClass.getSearchAdvanced(tool)
            if advanced: r.merge(advanced)
        elif (name == 'allSearch') or not name:
            # It is the search for every instance of p_modelClass
            r = Search('allSearch', container=modelClass)
            name = r.name
            # Take into account default params for the advanced search
            advanced = modelClass.getSearchAdvanced(tool)
            if advanced: r.merge(advanced)
        elif name == 'fromSearch':
            # It is the search for selecting a template for creating an object
            fromClass = req.fromClass
            sourceField = req.sourceField
            if sourceField:
                # We are creating an object from a Ref field
                id, fieldName = sourceField.split(':')
                o = tool.getObject(id)
                r = o.getField(fieldName).getAttribute(o, 'createVia')
            else:
                # We are creating a root object
                tool.model.classes.get(fromClass).getCreateVia(tool)
            initiator = initiators.TemplateInitiator(fromClass, req.formName,
                                                     req.insert, sourceField)
        elif ',' in name:
            # The search is defined in a field
            id, fieldName, mode = name.split(',')
            # Get the object with this "id". In the case of a Ref field with
            # link=popup, the initiator object can be a temp object.
            o = tool.getObject(id)
            field = o.getField(fieldName)
            if field.type == 'Ref':
                if mode.startswith('search*'):
                    # A search from Ref attribute "searches"
                    r = field.getSearch(o, mode[7:])
                else:
                    initiator = initiators.RefInitiator(o, field,
                                                        fieldName, mode)
                    r = field.getSelect(o)
            elif field.type == 'Computed':
                r = field.getSearch(o)
        elif name.isdigit():
            # The name of the search refers to a Query instance
            r = tool.getObject(name).getSearch()
        else:
            # Search among static or dynamic searches
            r = modelClass.getSearch(name, tool)
        # The search may not exist
        if not r: tool.raiseUnauthorized(tool.translate('search_broken'))
        # Return a UiSearch if required
        return r if not ui else UiSearch(r, tool, ctx, initiator, name)

    @classmethod
    def getRefInfo(class_, tool, info=None, nameOnly=True, sep=':'):
        '''When a search is triggered in the context of a Ref field, this method
           returns information about this reference: the source object and the
           Ref field.'''
        # It can happen in the following cases.
        # 1. A search is restricted to objects referenced through a Ref field
        # 2. A search is executed in the context of a Ref field with
        #    link="dropdown".
        if not info:
            # Try first to get the info from key "ref" in the request
            info = tool.req.ref
            if not info:
                # Try to get it from the "_ref" key within search criteria
                criteria = Criteria.readFromRequest(tool.H())
                if criteria and ('_ref' in criteria): info = criteria['_ref']
        if not info or (sep not in info): return None, None
        id, field = info.split(sep)
        o = tool.getObject(id)
        field = field if nameOnly else o.getField(field)
        return o, field

    @classmethod
    def encodeForReplay(class_, req, layout):
        '''Encodes in a string all the params in the p_req(uest) being required
           for re-playing a search.'''
        # If key "search" is present in the request, but empty, it represents
        # the special search named "allSearch".
        if ('search' in req) and not req.search:
            req.search = 'allSearch'
        if not req.search or (layout == 'cell'): return ''
        return ':'.join([(req[key] or '') for key in class_.replayParams])

    @classmethod
    def replay(class_, tool, searchParams):
        '''Replays the search whose parameters are in p_searchParams'''

        # In some contexts, a Search needs to be "replayed". For example,
        # imagine you run a search from the UI. While search results are shown
        # in the browser, a POD field is also available, to get results in a PDF
        # file. When the user asks to generate this file, the current search
        # must be replayed in order to deliver the resulting objects to the POD
        # field (in its context).
        #
        # Search parameters (p_searchParams) are in the format produced by
        # m_encodeForReplay. This method return resulting objects. The operation
        # is logged.
        # ~~~
        # Get search parameters as an object
        params = O()
        i = 0
        for part in searchParams.split(':', len(class_.replayParams) - 1):
            setattr(params, class_.replayParams[i], part)
            i += 1
        # Retrieve a Ref object when relevant
        if 'search*' in params.search:
            # Ref info may not be in the request, but may de deduced from the
            # search name.
            parts = params.search.split(',')
            info = '%s:%s' % (parts[0], parts[1])
        else:
            info = None
        refObject, refField = class_.getRefInfo(tool, info=info, nameOnly=False)
        # Get the Search instance
        modelClass = tool.model.classes.get(params.className)
        search = class_.get(params.search, tool, modelClass, None, ui=False)
        # Executing the search may take some time and potentially slow down
        # the system: log this action.
        tool.log(REPLAY_S % (modelClass.name, params.search))
        # Replay the search, but without restricting the number of results
        r = search.run(tool.H(), batch=False,
          sortBy=params.sortKey, sortOrder=params.sortOrder,
          filters=modelClass.getFilters(tool, params.filters),
          refObject=refObject, refField=refField)
        tool.log(REPLAY_OK % len(r))
        return r

    @classmethod
    def getCheckedInfo(class_, req):
        '''Gets the status of checked elements from the p_req(uest)'''

        # Those elements do not necessarily refer to search results: they could
        # also denote referred objects.
        # 
        # The method returns a tuple (ids, unchecked):
        # ----------------------------------------------------------------------
        #   "ids"     | is a dict of object IIDs;
        # "unchecked" | is True if these objects are unchecked, or False if
        #             | they are checked.
        # ----------------------------------------------------------------------

        ids = {int(v) for v in req.checkedIds.split(',')} if req.checkedIds \
                                                          else {}
        return ids, req.checkedSem == 'unchecked'

    @classmethod
    def keepCheckedResults(class_, req, objects):
        '''Among p_objects as retrieved via m_replay, keep only those being
           checked, according to m_getCheckedInfo.'''
        ids, unchecked = class_.getCheckedInfo(req)
        i = len(objects) - 1
        # Remove, from search results, unchecked objects
        while i >= 0:
            if unchecked: remove = objects[i].iid in ids
            else:         remove = objects[i].iid not in ids
            if remove:
                del objects[i]
            i -= 1

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  PXs
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    # Display results of a search whose name is in the request
    results = Px('''
     <div var="hook=uiSearch.getRootHook()" id=":hook">
      <script>:uiSearch.getCbJsInit(hook)</script>
      <x if="not mode.inField">::ui.Globals.getForms(tool)</x>
      <div align=":dright" if="len(resultModes) &gt; 1">
       <select name="px"
               onchange=":'switchResultMode(this, %s)' % q('searchResults')">
        <option for="mode in resultModes"
                value=":mode">:uiSearch.Mode.getText(mode, _)</option>
       </select>
      </div>
      <x>:uiSearch.pxResult</x>
     </div>''',

     js='''
       function switchResultMode(select, hook) {
         var mode = select.options[select.selectedIndex].value;
         askAjax(hook, null, {'resultMode': mode});
       }''',

     template=Template.px, hook='content')

    traverse['jresults'] = True
    @classmethod
    def jresults(class_, tool):
        '''Returns, as a JSON dict, in the context of a search whose name is in
           the request, the indexes of the siblings object of the one whose
           index is in the request.'''
        req = tool.req
        # The index of the current object
        index = int(req.index)
        # Get the search
        modelClass = tool.model.classes.get(req.className)
        search = class_.get(req.search, tool, modelClass, None, ui=False)
        # Determine the start of the batch search surrounding the object being
        # at "index".
        size = search.maxPerPage
        start = index - int(size / 2)
        if start < 0: start = 0
        # Perform the search
        ids = search.run(tool.H(), ids=True)
        r = {}
        i = start
        end = start + size
        while i < end:
            try:
                r[i] = ids[i]
            except IndexError:
                break
            i += 1
        tool.resp.setContentType('json')
        return sutils.getStringFrom(r, c='"')

    # Display results of a live search
    liveResults = Px('''
     <div var="search=uiSearch.search;
               filters=search.container.getFilters(tool);
               objects=search.run(handler,filters=filters).objects;
               io,ifield=search.getRefInfo(tool, pre, nameOnly=False, sep='_')"
          id=":'%s_LSResults' % pre">
      <p if="not objects">::_('query_no_result')</p>
      <div for="o in objects" style="padding: 3px 5px">
       <a href=":search.onSelectLiveResult(o, io, ifield)"
          var="title=o.getShownValue();
               content=Px.truncateValue(title, width=80)"
          title=":title">:content</a>
      </div>
      <!-- Go to the page showing all results -->
      <div if="objects and not io" align=":dright" style="padding: 3px">
       <a href=":'javascript:allLSResults(%s,%s)' % (q(pre), q(className))"
          class="lsAll">:_('search_results_all') + '...'</a>
      </div>
     </div>''')

    def onSelectLiveResult(self, o, io, ifield):
        '''Get the action to perform when the user click on a live search
           result.'''
        # If there is no *i*initiator *o*bject (io), redirect the user to p_o's
        # view page.
        if io is None: return o.url
        # Else, update the initiator field (io.ifield) in order to add p_o
        multi = 'true' if ifield.isMultiValued() else 'false'
        return "javascript:addLSResult('%s','%s','%s',%s)" % \
               (io.iid, ifield.name, o.iid, multi)

    # The advanced search form
    advanced = Px('''
     <x var="refInfo=req.ref;
             fields,css,js,gridder=class_.getSearchFields(tool, refInfo);
             layout='search'">

      <!-- Include type-specific CSS and JS -->
      <x>::ui.Includer.getSpecific(tool, css, js)</x>
      <script>var errors = null;</script>

      <!-- Search title -->
      <div class="pageTitle bottomSpace">
       <x>:_('%s_plural' % className)</x> 
       <img src=":url('arrowsA.svg')" class="iconS"
            style="transform: rotate(270deg)"/><x>:_('search_title')</x>
      </div>

      <!-- Form for searching objects of the requested class -->
      <form name="search" action=":'%s/Search/results' % tool.url"
            method="post">
       <input type="hidden" name="className" value=":className"/>
       <input type="hidden" name="source" value="searchForm"/>
       <input type="hidden" name="search" value="customSearch"/>
       <input if="refInfo" type="hidden" name="ref" value=":refInfo"/>

       <!-- The search fields -->
       <div style=":gridder.getContainerStyle() + ';margin-bottom: 1em'">
         <div for="field in fields"
            style=":'grid-column:span %d' %field.scolspan">:field.pxRender</div>
       </div>

       <!-- Submit button -->
       <input var="label=_('search_button');
                   css=ui.Button.getCss(label, small=False)"
              type="submit" class=":css" value=":label"
              style=":url('search.svg', bg='18px 18px')"/>
      </form>
     </x>''', template=Template.px, hook='content')

    # The live search PX. It requires variables:
    # --------------------------------------------------------------------------
    #  className | The name of the class whose instances will be searched
    # --------------------------------------------------------------------------
    #    pre     | The prefix used to determine identifiers for tags in use
    #            | within live search PXs. Indeed, if several live searches are
    #            | present on the same pages, these identifiers will allow to
    #            | distinguish them. For example, in the portlet, because there
    #            | is one live search per root class, pre == className. When
    #            | called from a Ref field with link='dropdown', pre is of the
    #            | form <Object IID>_<field name>, for example "666_managers".
    # --------------------------------------------------------------------------
    #   liveCss  | The CSS class(es) to assign to the main PX tag
    # --------------------------------------------------------------------------

    live = Px('''
     <div var="searchLabel=_('search_button')" class=":liveCss">
      <img src=":url('search.svg')" class="iconM"/>
      <div style="position: relative">
       <input type="text" size="14" name="w_searchable" autocomplete="off"
              id=":'%s_LSinput' % pre" class="inputSearch"
              title=":searchLabel" placeholder=":searchLabel"
              var="jsCall='onLiveSearchEvent(event,%s,%s,%s)' % \
                           (q(pre), q(className), q('auto'))"
              onkeyup=":jsCall" onfocus=":jsCall"
              onblur=":'onLiveSearchEvent(event,%s,%s,%s)' % \
                       (q(pre), q(className), q('hide'))"/>
       <!-- Dropdown containing live search results -->
       <div id=":'%s_LSDropdown' % pre" class="dropdown">
        <div id=":'%s_LSResults' % pre"></div>
       </div>
      </div>
     </div>''',

     css='''.lsSelected { background-color: #d9d7d9 }
            .lsRef { display: flex }
            .lsSearch { display: flex; margin-top: 5px; padding: |lsPadding|;
                        background-color: |lsBgColor| }
            .lsSearch input[type=text] { background-color: transparent;
                        border-bottom: |lsBottom| |lsBottomColor| }
            .lsSearch .dropdown { background-color: |lsrBgColor| }
            .lsAll { font-size: 80%; font-style: italic }''',

     # Live-search-related functions (LS)
     js='''
       function cleanKeywords(keywords) {
         return keywords.replace(/['",:]/g, '').trim();
       }

       function detectEventType(event) {
         /* After p_event occurred on a live search input field, must we trigger
            a search (a new char has been added), move up/down within the search
            results (key up/down has been pressed) or hide the dropdown
            (escape)? */
         if (event.type == 'focus') return 'search'
         switch (event.keyCode) {
           case 38: return 'up';
           case 40: return 'down';
           case 27: return 'hide'; // escape
           case 13: return 'go'; // cr
           case 37: break; // left
           case 39: break; // right
           default: return 'search';
         }
       }

       /* Function that selects the search result within the dropdown, after the
          user has pressed the 'up' od 'down' key (p_direction). */
       function selectLSResult(dropdown, direction){
         var results = dropdown.children[0].getElementsByTagName('div');
         if (results.length == 0) return;
         var j; // The index of the new element to select
         for (var i=0, len=results.length; i<len; i++) {
           if (results[i].className == 'lsSelected') {
             if (direction == 'up') {
               if (i > 0) j = i-1;
               else j = len-1;
             }
             else {
               if (i < (len-1)) j = i+1;
               else j = 0;
             }
             results[i].className = '';
             results[j].className = 'lsSelected';
             break;
           }
         }
         if (isNaN(j)) results[0].className = 'lsSelected';
       }

       // Function that allows to go to a selected search result
       function gotoLSLink(dropdown) {
         var a, results = dropdown.children[0].getElementsByTagName('div');
         for (var i=0, len=results.length; i<len; i++) {
           if (results[i].className == 'lsSelected') {
             a = results[i].children[0];
             if (a.href) window.location = a.href;
             else eval(a.onclick);
           }
         }
       }
    
       function hideLSDropdown(dropdown, timeout) {
         if (dropdown.style.display == 'none') return;
         if (!timeout) { dropdown.style.display = 'none'; return; }
         lsTimeout = setTimeout(function(){
           dropdown.style.display = 'none';}, 400);
       }

       // Retrieve all parameters for performing a live search
       function getLSParams(pre, className) {
         // Retrieve search keywords
         var keywords = document.getElementById(pre + '_LSinput').value;
         keywords = cleanKeywords(keywords);
         if (keywords[keywords.length-1] != '*') keywords = keywords + '*';         
         // Define base parameters
         var r = {'className':className, 'pre':pre,
                  'filters':'searchable:' + keywords};
         // Add parameters depending on the type of search to run
         if (pre.indexOf('_') == -1) {
           // A custom search triggered from a root class live search
           r['search'] = 'customSearch';
           r['source'] = 'searchForm';
         }
         else {
           // A live search triggered from a Ref field with link=dropdown
           r['search'] = pre.split('_').join(',') + ',add';
         }
         return r;
       }

       // Manage an p_event triggered from the LS input field
       function onLiveSearchEvent(event, pre, class_, action) {
         var dropdown = document.getElementById(pre + '_LSDropdown');
         if (lsTimeout) clearTimeout(lsTimeout);
         // Hide the dropdown if action is forced to 'hide'
         if (action == 'hide') { hideLSDropdown(dropdown, true); return; }
         // Detect if the dropdown must be shown or hidden
         var input = document.getElementById(pre + '_LSinput'),
             keywords = cleanKeywords(input.value || '');
         if (keywords.length > 2) {
           var eventType = detectEventType(event);
           if (!eventType) return;
           if (eventType == 'hide') { hideLSDropdown(dropdown, false); return;}
           if (eventType == 'go') { gotoLSLink(dropdown); return; }
           if (eventType == 'search') {
             // Trigger an Ajax search and refresh the dropdown content
             var params = getLSParams(pre, class_);
             lsTimeout = setTimeout(function() {
               askAjaxChunk(siteUrl + '/tool/Search/liveResults', 'GET', params,
                            pre + '_LSResults');
               dropdown.style.display = 'block';}, 400);
             }
           else { // Move up or down within results
             selectLSResult(dropdown, eventType);
           }
         }
         else { hideLSDropdown(dropdown, true); }
       }

       // Get all search results
       function allLSResults(pre, className) {
         var params = getLSParams(pre, className);
         post(siteUrl + '/tool/Search/results', params);
       }

       // Add the selected object to the initiator Ref field
       function addLSResult(oid, ifield, id, multi) {
         var hook = oid + '_' + ifield,
             url = siteUrl + '/' + oid;
         if (multi) {
           // Also get the already selected objects in ifield
           var sel = document.getElementById(ifield),
               selected=sel.selectedOptions, ids=[];
           for (var i=0; i<selected.length; i++) {
             ids.push(selected[i].value);
           }
           if (ids.indexOf(id) == -1) ids.push(id);
           ids = ids.join(',');
         }
         else { ids = id; }
         // Simulate objects selected via a popup
         askField(hook, url, 'edit', {'semantics': 'checked', 'selected': ids});
       }''')

    # An alias to batch results
    batchResults = UiSearch.pxResult
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''Counting Lines Of Code (LOC)'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from pathlib import Path

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class LanguageCounter:
    '''Counts lines of code for files written in a given language'''

    def __init__(self, name, spaces):
        # The name of the language
        self.name = name
        # The total number of analysed files
        self.numberOfFiles = 0
        # The number of empty lines within those files
        self.emptyLines = 0
        # The number of comment lines
        self.commentLines = 0
        # A code line is anything not being an empty or comment line
        self.codeLines = 0
        # The spaces to dump before printing anything to stdout
        self.spaces = spaces

    def numberOfLines(self):
        '''Computes the total number of lines within analysed files.'''
        return self.emptyLines + self.commentLines + self.codeLines

    def countAny(self, f, start, end, single=None):
        '''Analyses file p_f. p_start and p_end are delimiters for multi-line
           comments, while p_single is the symbol representing a single-line
           comment.'''
        inDoc = False
        firstDoc = False
        for line in f:
            stripped = line.strip()
            # Manage a single-line comment
            if not inDoc and single and line.startswith(single):
                self.commentLines += 1
                continue
            # Manage a comment
            if not inDoc and (start in line):
                if line.startswith(start):
                    self.commentLines += 1
                else:
                    self.codeLines += 1
                inDoc = True
                firstDoc = True
            if inDoc:
                if not firstDoc:
                    self.commentLines += 1
                if end in line:
                    inDoc = False
                firstDoc = False
                continue
            # Manage an empty line
            if not stripped:
                self.emptyLines += 1
            else:
                self.codeLines += 1

    def countJs(self, f): return self.countAny(f, '/*', '*/', single='//')
    countCss = countJs

    def countXml(self, f): return self.countAny(f, '<!--', '-->')

    # Doc separators in the Python language
    pythonDocSeps = ('"""', "'''")

    def isPythonDoc(self, line, start, isStart=False):
        '''Returns True if we find, in p_line, the start of a docstring (if
           p_start is True) or the end of a docstring (if p_start is False).
           p_isStart indicates if p_line is the start of the docstring.'''
        seps = LanguageCounter.pythonDocSeps
        if start:
            r = line.startswith(seps[0]) or line.startswith(seps[1])
        else:
            sepOnly = (line == seps[0]) or (line == seps[1])
            if sepOnly:
                # If the line contains the separator only, is this the start or
                # the end of the docstring ?
                r = not isStart
            else:
                r = line.endswith(seps[0]) or line.endswith(seps[1])
        return r

    def countPy(self, f):
        '''Analyses the Python file p_f'''
        # Are we in a docstring ?
        inDoc = False
        try:
            for line in f:
                stripped = line.strip()
                # Manage a line that is within a docstring
                inDocStart = False
                if not inDoc and self.isPythonDoc(stripped, start=True):
                    inDoc = True
                    inDocStart = True
                if inDoc:
                    self.commentLines += 1
                    if self.isPythonDoc(stripped, start=False, isStart=inDocStart):
                        inDoc = False
                    continue
                # Manage an empty line
                if not stripped:
                    self.emptyLines += 1
                    continue
                # Manage a comment line
                if line.startswith('#'):
                    self.commentLines += 1
                    continue
                # If we are here, we have a code line
                self.codeLines += 1
        except UnicodeDecodeError:
            print(self.spaces, '%s could not be read (unicode error).' % f.name)

    def countFile(self, path):
        '''Analyses file named p_path'''
        with path.open() as f:
            # Choose the method corresponding to the file extension
            method = 'count%s' % (path.suffix[1:].capitalize())
            getattr(self, method)(f)
        self.numberOfFiles += 1

    def printReport(self):
        '''Returns the analysis report as a string, only if there is at least
           one analysed line.'''
        lines = self.numberOfLines()
        if not lines: return
        commentRate = (self.commentLines / float(lines)) * 100.0
        blankRate = (self.emptyLines / float(lines)) * 100.0
        print(self.spaces,
              '%s: %d files, %d lines (%.0f%% comments, %.0f%% blank)' % \
              (self.name, self.numberOfFiles, lines, commentRate, blankRate))

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Counter:
    '''Counts and classifies the lines of code within a folder hierarchy'''

    # Paths containing these chunks will be excluded from the analysis
    defaultExclude = ('/.svn/', '/tmp/', '/temp/', '/jscalendar/', '/doc/')

    # Files corresponding to these languages will be part of the analysis
    languages = {'py': 'Python', 'xml': 'XML', 'css': 'CSS', 'js': 'Javascript'}

    def __init__(self, folder, exclude=None, spaces=''):
        # The spaces to dump before printing anything to stdout
        self.spaces = spaces
        # The base folder into which lines of code will be counted
        self.folder = Path(folder) if isinstance(folder, str) else folder
        # Collect here info about the analysed files ~{s_ext: LanguageCounter}~
        self.counters = { ext: LanguageCounter(name, spaces) \
                          for ext, name in Counter.languages.items()}
        # Which paths to exclude from the analysis ?
        self.exclude = list(Counter.defaultExclude)
        if exclude: self.exclude += exclude

    def printReport(self):
        '''Displays on stdout a small analysis report about self.folder'''
        total = 0
        for ext, counter in self.counters.items():
            if counter.numberOfFiles:
                counter.printReport()
                total += counter.numberOfLines()
        print(self.spaces, 'Total (including commented and blank): ***',
              total, '***')

    def mustExclude(self, path):
        '''Must p_path be excluded from the analysis ?'''
        for excl in self.exclude:
            if excl in path.as_posix(): return True

    def run(self):
        '''Start the count'''
        # Launch one analysis per language
        for ext, counter in self.counters.items():
            # Browse all files of this type in p_self.folder
            for path in self.folder.glob('**/*.%s' % ext):
                # Ignore paths to exclude
                if self.mustExclude(path): continue
                # Analyse the file in "path"
                self.counters[ext].countFile(path)
        self.printReport()
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import time, uuid
from appy.utils import Traceback

# POD-specific constants - - - - - - - - - - - - - - - - - - - - - - - - - - - -
XHTML_HEADINGS = ('h1', 'h2', 'h3', 'h4', 'h5', 'h6')
XHTML_LISTS    = ('ol', 'ul')

# "para" is a meta-tag representing p, div, blockquote or address
XHTML_PARA_TAGS = XHTML_HEADINGS + ('p', 'div', 'blockquote', 'address', 'para')
XHTML_INNER_TAGS      = ('b', 'i', 'u', 'em', 'span')
XHTML_UNSTYLABLE_TAGS = ('li', 'a')
XHTML_META_TAGS       = {'p': 'para', 'div': 'para',
                         'blockquote': 'para', 'address': 'para'}

# Cell-like tags (including "li")
XHTML_CELL_TAGS = ('th', 'td', 'li')

# Default output formats for POD templates, by template type
formatsByTemplate = {'.odt': ('pdf', 'doc', 'docx', 'odt'),
                     '.ods': ('xls', 'xlsx', 'ods')}

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Evaluator:
    '''Wrapper around the built-in Python function "eval"'''
    # When evaluating a Python expression, pod will use the built-in "eval"
    # function, like this:
    #             eval(expression, emptyContext, podContext)
    # "emptyContext" is the following variable: an empty dict.
    emptyContext = {}
    # "podContext" is the current pod context.

    @classmethod
    def run(class_, expression, context):
        '''Evaluates p_expression in this p_context'''
        # p_context may be a dict or an instance of appy.model.utils.Object. In
        # this latter case, although it implements dict-like methods, it cannot
        # be used as-is as local context. Indeed, it does not raise a KeyError
        # when a key lookup produces no result, but returns None instead. So,
        # for any non-local name, the "eval" function believes it has been found
        # in the local mapping, with value being None. This is why we unwrap the
        # __dict__ behind a context being an instance of
        # appy.model.utils.Object.
        context = context if isinstance(context, dict) else context.__dict__
        # Evaluate p_expression
        return eval(expression, class_.emptyContext, context)
        # Note that, after the first execution of function "eval", the function
        # adds, within "emptyDict", a key '__builtins__' containing all the
        # Python built-ins, similarly to the homonym entry in dict globals().

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class PodError(Exception):
    '''Base pod exception'''

    @staticmethod
    def dumpTraceback(buffer, tb, removeFirstLine):
        # If this error came from an exception raised by pod (p_removeFirstLine
        # is True), the text of the error may be very long, so we avoid having
        # it as error cause + in the first line of the traceback.
        linesToRemove = removeFirstLine and 3 or 2
        i = 0
        for tLine in tb.splitlines():
            i += 1
            if i > linesToRemove:
                buffer.write('<text:p>')
                try:
                    buffer.dumpContent(tLine)
                except UnicodeDecodeError:
                    buffer.dumpContent(tLine.decode('utf-8'))
                buffer.write('</text:p>')

    @staticmethod
    def dump(buffer, message, withinElement=None, removeFirstLine=False,
             dumpTb=True, escapeMessage=True):
        '''Dumps the error p_message in p_buffer or in r_ if p_buffer is None'''
        if withinElement:
            buffer.write('<%s>' % withinElement.OD.nsName)
            for subTag in withinElement.subTags:
                buffer.write('<%s>' % subTag.nsName)
        buffer.write('<office:annotation><dc:creator>POD</dc:creator>' \
          '<dc:date>%s</dc:date><text:p>' % time.strftime('%Y-%m-%dT%H:%M:%S'))
        if escapeMessage:
            buffer.dumpContent(message)
        else:
            buffer.write(message)
        buffer.write('</text:p>')
        if dumpTb:
            # We don't dump the traceback if it is an expression error (it is
            # already included in the error message).
            PodError.dumpTraceback(buffer, Traceback.get(), removeFirstLine)
        buffer.write('</office:annotation>')
        if withinElement:
            subTags = withinElement.subTags[:]
            subTags.reverse()
            for subTag in subTags:
                buffer.write('</%s>' % subTag.nsName)
            buffer.write('</%s>' % withinElement.OD.nsName)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def getUuid(removeDots=False, prefix='f'):
    '''Get a unique ID (ie, for images/documents to be imported into an ODT
       document).'''
    r = uuid.uuid4().hex
    if removeDots: r = r.replace('.', '')
    return '%s%s' % (prefix, r)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-
# I need a class.
class D:
    def getAt1(self):
        return '''
<p>Notifia</p>
<ol>
  <li>Een</li>
  <li>Een</li>
  <li>Een</li>
  <li class="podItemKeepWithNext">Keep with next, without style mapping.</li>
  <li>Een</li>
  <li>Een</li>
  <li>Een</li>
  <li>Een</li>
  <li>Een</li>
  <li>Een</li>
  <li>Een</li>
  <ul><li class="pmItemKeepWithNext">This one has 'keep with next'</li>
        <li>Hello</li>
          <ol><li>aaaaaaaaaa aaaaaaaaaaaaaa</li>
              <li>aaaaaaaaaa aaaaaaaaaaaaaa</li>
              <li>aaaaaaaaaa aaaaaaaaaaaaaa</li>
              <li class="pmItemKeepWithNext">This one has 'keep with next'</li>
          </ol>
  </ul>
  <li>Een</li>
  <li>Een</li>
  <li>Een</li>
  <li>Een</li>
  <li>Een</li>
  <li>Een</li>
  <li class="pmItemKeepWithNext">This one has 'keep with next'</li>
</ol>
<ul>
  <li>Un</li>
  <li>Deux</li>
  <li>Trois</li>
  <li>Quatre</li>
  <li class="pmItemKeepWithNext">VCinq (this one has 'keep with next')</li>
  <li class="pmItemKeepWithNext">Six (this one has 'keep with next')</li>
  <li class="pmItemKeepWithNext">Sept (this one has 'keep with next')</li>
</ul>'''
dummy = D()

xhtml1 = '''
<p>First line</p>
<p>Second line: must be keepWithNext</p>
'''

xhtml2 = '''
<p>First line</p>
<ul>
 <li>1</li>
 <li>2</li>
 <li>This should be KWN</li>
</ul>
'''

xhtml3 = '''
<p>First line</p>
<ol>
 <li>This NB should be KWN</li>
</ol>
'''

# "Keep with next" with the ParaKWN class explicitly used
xhtml4 = '''
<p>First para</p>
<p class="ParaKWN">Hello1</p>
<p class="ParaKWN" style="text-align:justify">
 <span style="font-size:10.0pt">Hello.</span></p>
'''

# Idem, case #2
xhtml5 = '''
<p>TEST KWN #2</p>
<p style="text-align:justify">This line generates a style that must not be reused</p>
<p class="ParaKWN" style="text-align:justify">KWN</p>
'''

# "Keep with next" is set to 20: the 2 last lines must be touched
xhtml6 = '''
<p>TEST KWN <br/> #3</p>
<div>TEST KWN #3</div>
<ol><li>TEST KWN #3</li></ol>
'''

# "Keep with next" in the middle of a table (kwn=30)
xhtml7 = '''
<table>
 <tr>
   <td>Aaa</td><td>Bbb</td>
 </tr>
 <tr>
   <td>A a</td><td>B gtrezazetink</td>
 </tr>
</table>
<p>Last line Last li 20</p>
'''

# "Keep with next" with a table whose attribute "style" contains an entity
xhtml8 = '''
<table style="border:undefined; font-family:&quot;Calibri&quot;; font-size:x-small">
 <tr><td>A</td><td>B</td></tr>
</table>'''



# ------------------------------------------------------------------------------
import os, re, time, copy
from utils import produceNiceMessage

# ------------------------------------------------------------------------------
poHeader = '''msgid ""
msgstr ""
"Project-Id-Version: %s\\n"
"POT-Creation-Date: %s\\n"
"MIME-Version: 1.0\\n"
"Content-Type: text/plain; charset=utf-8\\n"
"Content-Transfer-Encoding: 8bit\\n"
"Plural-Forms: nplurals=1; plural=0\\n"
"Language-code: %s\\n"
"Language-name: %s\\n"
"Preferred-encodings: utf-8 latin1\\n"
"Domain: %s\\n"
%s

'''
fallbacks = {'en': 'en-us en-ca',
             'fr': 'fr-be fr-ca fr-lu fr-mc fr-ch fr-fr'}

# Default values for i18n labels whose ids are not fixed.
CONFIG = "Configuration panel for product '%s'"
CONFIRM = 'Are you sure ?'

# ------------------------------------------------------------------------------
class PoMessage:
    '''Represents a i18n message (po format).'''
    def __init__(self, id, msg, default, fuzzy=False, comments=[],
                 niceDefault=False):
        self.id = id
        self.msg = msg
        self.default = default
        if niceDefault: self.produceNiceDefault()
        self.fuzzy = fuzzy # True if the default value has changed in the pot
        # file: the msg in the po file needs to be translated again.
        self.comments = comments

    def update(self, newMsg, isPot, language):
        '''Updates me with new values from p_newMsg. If p_isPot is True (which
           means that the current file is a pot file), I do not care about
           filling self.msg.'''
        if isPot:
            self.msg = ""
            if not self.default:
                self.default = newMsg.default
                # It means that if the default message has changed, we will not
                # update it in the pot file. We will always keep the one that
                # the user may have changed in the pot file. We will write a
                # default message only when no default message is defined.
        else:
            # newMsg comes from a pot file or from a base po file (like a
            # standard Appy po file). We must update the corresponding
            # message in the current po file.
            oldDefault = self.default
            if self.default != newMsg.default:
                # The default value has changed in the pot file
                oldDefault = self.default or ''
                self.default = newMsg.default
                self.fuzzy = False
                if self.msg.strip():
                    self.fuzzy = True
                    # We mark the message as "fuzzy" (=may need to be rewritten
                    # because the default value has changed) only if the user
                    # has already entered a message. Else, this has no sense to
                    # rewrite the empty message.
                    if not oldDefault.strip():
                        # This is a strange case: the old default value did not
                        # exist. Maybe was this PO file generated from some
                        # tool, but simply without any default value. So in
                        # this case, we do not consider the label as fuzzy.
                        self.fuzzy = False
            # If p_newMsg contains a message, and no message is defined for
            # self, copy it.
            if newMsg.msg and not self.msg:
                self.msg = newMsg.msg
            # For english, the the default value from a pot file can be used as
            # value for the po file.
            if (language == 'en'):
                if not self.msg:
                    # Put the default message into msg for english
                    self.msg = self.default
                if self.fuzzy and (self.msg == oldDefault):
                    # The message was equal to the old default value. It means
                    # that the user did not change it, because for English we
                    # fill by default the message with the default value (see
                    # code just above). So in this case, the message was not
                    # really fuzzy.
                    self.fuzzy = False
                    self.msg = self.default

    def produceNiceDefault(self):
        '''Transforms self.default into a nice msg.'''
        self.default = produceNiceMessage(self.default)

    def generate(self):
        '''Produces myself as I must appear in a po(t) file.'''
        res = ''
        for comment in self.comments:
            res += comment + '\n'
        if self.default != None:
            res = '#. Default: "%s"\n' % self.default
        if self.fuzzy:
            res += '#, fuzzy\n'
        res += 'msgid "%s"\n' % self.id
        res += 'msgstr "%s"\n' % self.msg
        return res

    def __repr__(self):
        return '<i18n msg id="%s", msg="%s", default="%s">' % \
               (self.id, self.msg, self.default)

    def getMessage(self):
        '''Returns self.msg, but with some replacements.'''
        return self.msg.replace('<br/>', '\n').replace('\\"', '"')

class PoMessages:
    '''A list of po messages under construction.'''
    def __init__(self):
        # The list of messages
        self.messages = []
        # A dict of message ids, useful for efficiently checking if an id is
        # already in the list or not.
        self.ids = {}

    def append(self, id, default, nice=True):
        '''Creates a new PoMessage and adds it to self.messages. If p_nice is
           True, it produces a nice default value for the message.'''
        # Avoir creating duplicate ids
        if id in self.ids: return
        message = PoMessage(id, '', default, niceDefault=nice)
        self.messages.append(message)
        self.ids[id] = True

    def get(self): return self.messages

class PoHeader:
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def generate(self):
        '''Generates the representation of myself into a po(t) file.'''
        return '"%s: %s\\n"\n' % (self.name, self.value)

class PoFile:
    '''Represents a i18n file.'''
    def __init__(self, fileName):
        self.fileName = fileName
        self.isPot = fileName.endswith('.pot')
        self.messages = [] # Ordered list of messages
        self.messagesDict = {} # Dict of the same messages, indexed by msgid
        self.headers = []
        self.headersDict = {}
        # Get application name, domain name and language from fileName
        self.applicationName = ''
        self.language = ''
        self.domain = ''
        baseName = os.path.splitext(os.path.basename(fileName))[0]
        elems = baseName.split('-')
        if self.isPot:
            if len(elems) == 1:
                self.applicationName = self.domain = elems[0]
            else:
                self.applicationName, self.domain = elems
        else:
            if len(elems) == 1:
                self.applicationName = self.domain = ''
                self.language = elems[0]
            elif len(elems) == 2:
                self.applicationName = self.domain = elems[0]
                self.language = elems[1]
            else:
                self.applicationName, self.domain, self.language = elems
        self.generated = False # Will be True during the generation process,
        # once this file will have been generated.

    def addMessage(self, newMsg, needsCopy=True):
        if needsCopy:
            res = copy.copy(newMsg)
        else:
            res = newMsg
        self.messages.append(res)
        self.messagesDict[res.id] = res
        return res

    def addHeader(self, newHeader):
        self.headers.append(newHeader)
        self.headersDict[newHeader.name] = newHeader

    def update(self, newMessages, removeNotNewMessages=False,
               keepExistingOrder=True):
        '''Updates the existing messages with p_newMessages.
           If p_removeNotNewMessages is True, all messages in self.messages
           that are not in newMessages will be removed, excepted if they start
           with "custom_". If p_keepExistingOrder is False, self.messages will
           be sorted according to p_newMessages. Else, newMessages that are not
           yet in self.messages will be appended to the end of self.messages.'''
        # First, remove not new messages if necessary
        newIds = [m.id for m in newMessages]
        removedIds = []
        if removeNotNewMessages:
            i = len(self.messages)-1
            while i >= 0:
                oldId = self.messages[i].id
                if not oldId.startswith('custom_') and (oldId not in newIds):
                    del self.messages[i]
                    del self.messagesDict[oldId]
                    removedIds.append(oldId)
                i -= 1
        if keepExistingOrder:
            # Update existing messages and add inexistent messages to the end.
            for newMsg in newMessages:
                if self.messagesDict.has_key(newMsg.id):
                    msg = self.messagesDict[newMsg.id]
                else:
                    msg = self.addMessage(newMsg)
                msg.update(newMsg, self.isPot, self.language)
        else:
            # Keep the list of all old messages not being in new messages.
            # We will append them at the end of the new messages.
            notNewMessages = [m for m in self.messages if m.id not in newIds]
            del self.messages[:]
            for newMsg in newMessages:
                if self.messagesDict.has_key(newMsg.id):
                    msg = self.messagesDict[newMsg.id]
                    self.messages.append(msg)
                else:
                    msg = self.addMessage(newMsg)
                msg.update(newMsg, self.isPot, self.language)
            # Append the list of old messages to the end
            self.messages += notNewMessages
        return removedIds

    def generateHeaders(self, f):
        if not self.headers:
            creationTime = time.strftime("%Y-%m-%d %H:%M-%S", time.localtime())
            fb = ''
            if not self.isPot:
                # I must add fallbacks
                if fallbacks.has_key(self.language):
                    fb = '"X-is-fallback-for: %s\\n"' % fallbacks[self.language]
            f.write(poHeader % (self.applicationName, creationTime,
                                self.language, self.language, self.domain, fb))
        else:
            # Some headers were already found, we dump them as is
            f.write('msgid ""\nmsgstr ""\n')
            for header in self.headers:
                f.write(header.generate())
            f.write('\n')

    def generate(self):
        '''Generates the corresponding po or pot file.'''
        folderName = os.path.dirname(self.fileName)
        if not os.path.exists(folderName):
            os.makedirs(folderName)
        f = file(self.fileName, 'w')
        self.generateHeaders(f)
        for msg in self.messages:
            f.write(msg.generate())
            f.write('\n')
        f.close()
        self.generated = True

    def getPoFileName(self, language):
        '''Gets the name of the po file that corresponds to this pot file and
           the given p_language.'''
        if self.applicationName == self.domain:
            res = '%s-%s.po' % (self.applicationName, language)
        else:
            res = '%s-%s-%s.po' % (self.applicationName, self.domain, language)
        return res

    def getCustomMessages(self):
        '''Returns, the list of messages from self.messages whose ID starts with
           "custom_".'''
        return [m for m in self.messages if m.id.startswith('custom_')]

class PoParser:
    '''Allows to parse a i18n file. The result is produced in self.res as a
       PoFile instance.'''
    def __init__(self, fileName):
        self.res = PoFile(fileName)

    # Regular expressions for msgIds, msgStrs and default values.
    re_default = re.compile('#\.\s+Default\s*:\s*"(.*)"')
    re_fuzzy = re.compile('#,\s+fuzzy')
    re_id = re.compile('msgid\s+"(.*)"')
    re_msg = re.compile('msgstr\s+"(.*)"')

    def parse(self):
        '''Parses all i18n messages in the file, stores it in
           self.res.messages and returns self.res.'''
        f = file(self.res.fileName)
        # Currently parsed values
        msgDefault = msgFuzzy = msgId = msgStr = None
        comments = []
        # Walk every line of the po(t) file
        for line in f:
            lineContent = line.strip()
            if lineContent and (not lineContent.startswith('"')):
                r = self.re_default.match(lineContent)
                if r:
                    msgDefault = r.group(1)
                else:
                    r = self.re_fuzzy.match(lineContent)
                    if r:
                        msgFuzzy = True
                    else:
                        r = self.re_id.match(lineContent)
                        if r:
                            msgId = r.group(1)
                        else:
                            r = self.re_msg.match(lineContent)
                            if r:
                                msgStr = r.group(1)
                            else:
                                if lineContent.startswith('#'):
                                    comments.append(lineContent.strip())
                if msgStr != None:
                    if not ((msgId == '') and (msgStr == '')):
                        poMsg = PoMessage(msgId, msgStr, msgDefault, msgFuzzy,
                                          comments)
                        self.res.addMessage(poMsg)
                    msgDefault = msgFuzzy = msgId = msgStr = None
                    comments = []
            if lineContent.startswith('"'):
                # It is a header value
                name, value = lineContent.strip('"').split(':', 1)
                if value.endswith('\\n'):
                    value = value[:-2]
                self.res.addHeader(PoHeader(name.strip(), value.strip()))
        f.close()
        return self.res
# ------------------------------------------------------------------------------



'''Management of the Appy authentication cookie'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import base64, urllib.parse

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Cookie:
    '''Represents the Appy authentication cookie, carrying user credentials. It
       will be set and read by an Appy app for authenticating a user.'''

    # Name of the Appy authentication cookie
    name = 'AppyAuth'

    @classmethod
    def read(class_, handler, onResponse=False):
        '''Returns the tuple (login, password, ctx) read from the authentication
           cookie read from the request. If no user is logged, its returns
           (None, None, None).'''
        # If p_onResponse, we get the cookie that has already been set on the
        # response, instead of the one coming from the request.
        if not onResponse:
            cookie = handler.req[Cookie.name]
        else:
            cookie = handler.resp.headers['Cookies'][Cookie.name]
        if not cookie: return None, None, None
        unquoted = urllib.parse.unquote(cookie).encode('utf-8')
        cookieValue = base64.decodestring(unquoted).decode('utf-8')
        if ':' not in cookieValue: return None, None, None
        # Extract the context from the cookieValue
        r, context = cookieValue.rsplit(':', 1)
        # Extract login and password
        login, password = r.split(':', 1)
        return login, password, context

    @classmethod
    def write(class_, handler, login, password, ctx=None):
        '''Encode p_login, p_password and p_ctx into the authentication
           cookie.'''
        r = '%s:%s:%s' % (login, password, ctx or '')
        r = base64.encodestring(r.encode('utf-8')).rstrip()
        handler.resp.setCookie(Cookie.name, urllib.parse.quote(r))

    @classmethod
    def update(class_, handler, ctx, onResponse=False):
        '''Updates the authentication context within the Appy authentication
           cookie. If p_onResponse is True, we get the cookie from the response
           headers instead of the request object.'''
        login, password, oldCtx = Cookie.read(handler, onResponse=onResponse)
        if login is None:
            # There is no cookie. When SSO is enabled, cookies are normally not
            # used, excepted if an authentication context comes into play. In
            # that case, the cookie is only used to store this context. In order
            # to represent this fact, set, in the cookie, a special value for
            # the user login:
            #                          _s_s_o_
            #
            # Such a cookie means: "this is a special cookie, only used to store
            # the authentication context. Check SSO to get the user login".
            login = '_s_s_o_'
        Cookie.write(handler, login, password, ctx)

    @classmethod
    def updatePassword(class_, handler, password):
        '''Updates the cookie with a new p_password'''
        login, oldPassword, ctx = Cookie.read(handler)
        Cookie.write(handler, login, password, ctx)

    @classmethod
    def disable(class_, handler):
        '''Disables the authentication cookie'''
        # Do it only if the cookie exists
        if Cookie.name in handler.req:
            handler.resp.setCookie(Cookie.name, 'deleted')
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

# ------------------------------------------------------------------------------
from appy.px import Px
from appy.fields import Field
from appy.gen.layout import Layouts
from appy.shared import utils as sutils
from appy.gen.indexer import ValidValue

# ------------------------------------------------------------------------------
emptyTuple = ()
RADIOS_KO = "This field can't be rendered as radio buttons because it may " \
            "contain several values (max multiplicity is higher than 1)."
CBS_KO    = "This field can't be rendered as checkboxes because it can " \
            "only contain a single value (max multiplicity is 1)."
IEDIT_KO  = 'It is currently not possible to inline-edit a multi-valued ' \
            'Select field.'

# ------------------------------------------------------------------------------
class Selection:
    '''If you want to have dynamically computed possible values for a Select
       field, use a Selection instance.'''

    def __init__(self, methodName):
        # p_methodName must be the name of a method that will be called every
        # time Appy will need to get the list of possible values for the related
        # field. It must correspond to an instance method of the class defining
        # the related field. This method accepts no argument and must return a
        # list (or tuple) of pairs (lists or tuples): (id, text), where "id" is
        # one of the possible values for the field, and "text" is the value as
        # will be shown on the screen. You can use self.translate within this
        # method to produce an i18n version of "text" if needed.
        self.methodName = methodName

    def getText(self, obj, value, field, language=None):
        '''Gets the text that corresponds to p_value'''
        if language:
            withTranslations = language
        else:
            withTranslations = True
        vals = field.getPossibleValues(obj, ignoreMasterValues=True,\
                                       withTranslations=withTranslations)
        for v, text in vals:
            if v == value: return text
        return value

# ------------------------------------------------------------------------------
class Select(Field):
    '''Field allowing to choose a value among a list of possible values. Each
       value is represented and stored as a string.'''

    pxView = Px('''
     <!-- No value at all -->
     <span if="not value" class="smaller">-</span>
     <!-- A single value -->
     <x if="value and not isMultiple">::field.getInlineEditableValue(obj, \
                                          value, layoutType, name=name)</x>
     <!-- Several values -->
     <ul if="value and isMultiple"><li for="sv in value"><i>::sv</i></li></ul>
     <!-- If this field is a master field -->
     <input type="hidden" if="masterCss" class=":masterCss" value=":rawValue"
            name=":name" id=":name"/>''')

    # More compact representation on the cell layout
    pxCell = Px('''
     <x var="multiple=value and isMultiple">
      <x if="multiple">:', '.join(value)</x>
      <x if="not multiple">:field.pxView</x>
     </x>''')

    pxEdit = Px('''
     <x var="isSelect=field.render == 'select';
             possibleValues=field.getPossibleValues(zobj, \
               withTranslations=True, withBlankValue=isSelect);
             charsWidth=field.getWidthInChars(False)">
     <select if="isSelect" name=":name" id=":name" class=":masterCss"
       multiple=":isMultiple" onchange=":field.getOnChange(zobj, layoutType)"
       size=":field.getSelectSize(False, isMultiple)"
       style=":field.getSelectStyle(False, isMultiple)">
      <option for="val, text in possibleValues" value=":val"
              selected=":field.isSelected(zobj, name, val, rawValue)"
              title=":text">:ztool.truncateValue(text, charsWidth)</option>
     </select>
     <x if="not isSelect">
      <div for="val, text in possibleValues">
       <input type=":field.render" name=":name" id=":val" value=":val"
              class=":masterCss" onchange=":field.getOnChange(zobj, layoutType)"
              checked=":field.isSelected(zobj, name, val, rawValue)"/>
       <label lfor=":val" class="subLabel">:text</label>
      </div>
     </x>
     <script if="hostLayout" var2="x=zobj.setLock(user, field=field)">:\
       'prepareForAjaxSave(%s,%s,%s,%s)' % \
        (q(name),q(obj.id),q(obj.url),q(hostLayout))</script></x>''')

    # On the search form, show a multi-selection widget with a "AND/OR" selector
    pxSearch = Px('''
     <!-- The "and" / "or" radio buttons -->
     <x if="field.multiplicity[1] != 1"
        var2="operName='o_%s' % name;
              orName='%s_or' % operName;
              andName='%s_and' % operName">
      <input type="radio" name=":operName" id=":orName"
             checked="checked" value="or"/>
      <label lfor=":orName">:_('search_or')</label>
      <input type="radio" name=":operName" id=":andName" value="and"/>
      <label lfor=":andName">:_('search_and')</label><br/>
     </x>

     <!-- The list of values -->
     <select var="preSelected=field.sdefault;
                  charsWidth=field.getWidthInChars(True)"
       name=":widgetName" multiple="multiple"
       size=":field.getSelectSize(True, True)"
       style=":field.getSelectStyle(True, True)"
       onchange=":field.getOnChange(ztool, 'search', className)">
      <option for="val, text in field.getPossibleValues(ztool, \
            withTranslations=True, withBlankValue=False, className=className)"
        selected=":val in preSelected" value=":val"
        title=":text">:ztool.truncateValue(text, charsWidth)</option>
     </select><br/>''')

    # Widget for filtering object values on search results
    pxFilter = Px('''
     <select var="name=field.name;
                  filterId='%s_%s' % (mode.ajaxHookId, name);
                  charsWidth=field.getWidthInChars(True)"
        id=":filterId" name=":filterId" class="discreet"
          onchange=":'askBunchFiltered(%s,%s)' % (q(mode.ajaxHookId), q(name))">
      <option for="val, text in field.getPossibleValues(ztool, \
                withTranslations=True, withBlankValue='forced', \
                blankLabel='everything', className=className)"
       selected=":(name in mode.filters) and (mode.filters[name] == val)"
       value=":val" title=":text">:ztool.truncateValue(text,charsWidth)</option>
     </select>''')

    def __init__(self, validator=None, multiplicity=(0,1), default=None,
      defaultOnEdit=None, show=True, page='main', group=None, layouts=None,
      move=0, indexed=False, mustIndex=True, indexValue=None, searchable=False,
      specificReadPermission=False, specificWritePermission=False, width=None,
      height=None, maxChars=None, colspan=1, master=None, masterValue=None,
      focus=False, historized=False, mapping=None, generateLabel=None,
      label=None, sdefault='', scolspan=1, swidth=None, sheight=None,
      persist=True, inlineEdit=False, view=None, cell=None, edit=None, xml=None,
      translations=None, noValueLabel='choose_a_value', render='select'):
        # When choosing a value in a select widget, the entry representing no
        # value is translated according the label defined in attribute
        # "noValueLabel". The default one is something like "[ choose ]", but if
        # you prefer a less verbose version, you can use "no_value" that simply
        # displays a dash, or your own label.
        self.noValueLabel = noValueLabel
        # A Select field, is, by default, rendered as a HTML select widget, but
        # there are alternate render modes. Here are all the possible render
        # modes.
        # ----------------------------------------------------------------------
        #  render    | The field is rendered as...
        # ----------------------------------------------------------------------
        # "select"   | an HTML select widget, rendered as a dropdown list if
        #            | max multiplicity is 1 or as a selection box if several
        #            | values can be chosen.
        # ----------------------------------------------------------------------
        # "radio"    | radio buttons. This mode is valid for fields with a max
        #            | multiplicity being 1.
        # ----------------------------------------------------------------------
        # "checkbox" | checkboxes. This mode is valid for fields with a max
        #            | multiplicity being higher than 1.
        # ----------------------------------------------------------------------
        self.render = render
        # Call the base constructor
        Field.__init__(self, validator, multiplicity, default, defaultOnEdit,
          show, page, group, layouts, move, indexed, mustIndex, indexValue,
          searchable, specificReadPermission, specificWritePermission, width,
          height, maxChars, colspan, master, masterValue, focus, historized,
          mapping, generateLabel, label, sdefault, scolspan, swidth, sheight,
          persist, inlineEdit, view, cell, edit, xml, translations)
        # self.sdefault must be a list of value(s)
        self.sdefault = sdefault or []
        # Default width, height and maxChars
        if width is None:
            self.width = 30
        if height is None:
            self.height = 4
        # Define the filter PX when appropriate
        if self.indexed:
            self.filterPx = 'pxFilter'
        self.swidth = self.swidth or self.width
        self.sheight = self.sheight or self.height
        self.checkParameters()

    def checkParameters(self):
        '''Ensure coherence between parameters'''
        # Valid render modes depend on multiplicities
        multiple = self.isMultiValued()
        if multiple and (self.render == 'radio'):
            raise Exception(RADIOS_KO)
        if not multiple and (self.render == 'checkbox'):
            raise Exception(CBS_KO)
        # It is currently not possible to inline-edit a multi-valued field
        if multiple and self.inlineEdit:
            raise Exception(IEDIT_KO)

    def isSelected(self, obj, name, possibleValue, dbValue):
        '''When displaying a Select field, must the p_possibleValue appear as
           selected? p_name is given and used instead of field.name because it
           may contain a row number from a field within a List field.'''
        req = obj.REQUEST
        # Get the value we must compare (from request or from database)
        if req.has_key(name):
            compValue = req.get(name)
        else:
            compValue = dbValue
        # Compare the value
        if type(compValue) in sutils.sequenceTypes:
            return possibleValue in compValue
        return possibleValue == compValue

    def getDefaultLayouts(self):
        '''Returns the default layouts for this type'''
        return self.inGrid() and Layouts.Select.g or Layouts.b

    def getValue(self, obj, name=None, layout=None, noListIfSingleObj=False):
        value = Field.getValue(self, obj, name, layout)
        if not value:
            if self.isMultiValued(): return emptyTuple
            else: return value
        if isinstance(value, basestring) and self.isMultiValued():
            value = [value]
        elif isinstance(value, tuple):
            value = list(value)
        return value

    def getTranslatedValue(self, obj, value, language=None):
        '''Get the translated text for p_value, when p_value is one of p_self's
           authorized "static" values according to its p_self.validator (being a
           list or tuple o values).'''
        if self.translations:
            # Translations are available via a FieldTranslations instance
            language = language or obj.getUserLanguage()
            r = self.translations.get('value', language, value=value)
        else:
            r = obj.translate('%s_list_%s' % (self.labelId, value))
        return r

    def getFormattedValue(self, obj, value, layoutType='view',
                          showChanges=False, language=None):
        '''Select-specific value formatting'''
        # Return an empty string if there is no p_value
        if Field.isEmptyValue(self, obj, value) and not showChanges: return ''
        if isinstance(self.validator, Selection):
            # Value(s) come from a dynamic vocabulary
            val = self.validator
            if self.isMultiValued():
                r = [val.getText(obj, v, self, language) for v in value]
            else:
                r = val.getText(obj, value, self, language)
        else:
            # Values come from a fixed vocabulary whose texts are in i18n files
            _ = self.getTranslatedValue
            if self.isMultiValued():
                r = [_(obj, v, language=language) for v in value]
            else:
                r = _(obj, value, language=language)
        return r

    def getPossibleValues(self, obj, withTranslations=False,
                          withBlankValue=False, blankLabel=None, className=None,
                          ignoreMasterValues=False):
        '''Returns the list of possible values for this field (only for fields
           with self.isSelect=True). If p_withTranslations is True, instead of
           returning a list of string values, the result is a list of tuples
           (s_value, s_translation). Moreover, p_withTranslations can hold a
           given language: in this case, this language is used instead of the
           user language. If p_withBlankValue is True, a blank value is
           prepended to the list, excepted if the type is multivalued. Used in
           combination with p_withTranslations being True, the i18n label for
           translating the blank value is given in p_blankLabel. If p_className
           is given, p_obj is the tool and, if we need an instance of
           p_className, we will need to use p_obj.executeQuery to find one.'''
        # Get the user language for translations, from p_withTranslations
        lg = isinstance(withTranslations, str) and withTranslations or None
        req = obj.REQUEST
        master = self.master
        aobj = obj.appy()
        if not ignoreMasterValues and master and callable(self.masterValue):
            # This field is an ajax-updatable slave. Get the master value...
            if master.valueIsInRequest(obj, req):
                # ... from the request if available
                requestValue = master.getRequestValue(obj)
                masterValues = master.getStorableValue(obj, requestValue,
                                                       complete=True)
            elif not className:
                # ... or from the database if we are editing an object
                masterValues = master.getValue(obj)
            else:
                # We don't have any master value
                masterValues = None
            # Get possible values by calling self.masterValue
            if masterValues:
                values = self.masterValue(aobj, masterValues)
            else:
                values = []
            # Manage parameter p_withTranslations
            if not withTranslations: res = values
            else:
                res = []
                for v in values:
                    res.append((v, self.getFormattedValue(obj,v,language=lg)))
        else:
            # Get the possible values from attribute "validator"
            if isinstance(self.validator, Selection):
                # We need to call self.methodName for getting the (dynamic)
                # values. If methodName begins with _appy_, it is a special Appy
                # method: we will call it on the Mixin (=p_obj) directly. Else,
                # it is a user method: we will call it on the wrapper
                # (p_obj.appy()). Some args can be hidden into p_methodName,
                # separated with stars, like in this example: method1*arg1*arg2.
                # Only string params are supported.
                methodName = self.validator.methodName
                # Unwrap parameters if any.
                if methodName.find('*') != -1:
                    elems = methodName.split('*')
                    methodName = elems[0]
                    args = elems[1:]
                else:
                    args = ()
                # On what object must we call the method that will produce the
                # values?
                if methodName.startswith('tool:'):
                    obj = obj.getTool()
                    methodName = methodName[5:]
                else:
                    # We must call on p_obj. But if we have something in
                    # p_className, p_obj is the tool and not an instance of
                    # p_className as required. So find such an instance.
                    if className:
                        brains = obj.executeQuery(className, maxResults=1,
                                                  brainsOnly=True)
                        if brains:
                            obj = brains[0]._unrestrictedGetObject()
                # Do we need to call the method on the object or on the wrapper?
                if methodName.startswith('_appy_'):
                    exec 'res = obj.%s(*args)' % methodName
                else:
                    exec 'res = obj.appy().%s(*args)' % methodName
                if not withTranslations: res = [v[0] for v in res]
                elif isinstance(res, list): res = res[:]
            else:
                # The list of (static) values is in self.validator
                res = []
                for value in self.validator:
                    if withTranslations:
                        text = self.getTranslatedValue(obj, value, language=lg)
                        res.append((value, text))
                    else:
                        res.append(value)
        if (withBlankValue == 'forced') or \
           (withBlankValue and not self.isMultiValued()):
            # Create the blank value to insert at the beginning of the list
            if withTranslations:
                label = blankLabel or self.noValueLabel
                blankValue = ('', obj.translate(label, language=lg))
            else:
                blankValue = ''
            # Insert the blank value in the result
            if isinstance(res, tuple):
                res = (blankValue,) + res
            else:
                res.insert(0, blankValue)
        return res

    def validateValue(self, obj, value):
        '''Ensure p_value is among possible values'''
        possibleValues = self.getPossibleValues(obj, ignoreMasterValues=True)
        if isinstance(value, basestring):
            error = value not in possibleValues
        else:
            error = False
            for v in value:
                if v not in possibleValues:
                    error = True
                    break
        if error: return obj.translate('bad_select_value')

    def getStorableValue(self, obj, value):
        '''Get a multivalued value when appropriate'''
        if value and self.isMultiValued() and \
           (type(value) not in sutils.sequenceTypes):
            value = [value]
        return value

    def getIndexType(self): return 'ListIndex'

    def getIndexValue(self, obj, forSearch=False):
        '''Take care of empty values'''
        r = Field.getIndexValue(self, obj, forSearch)
        return ValidValue.forString(r, forSearch)
# ------------------------------------------------------------------------------



'''Descriptor classes defined in this file are "intermediary" classes that
   gather, from the user application, information about found gen- or workflow-
   classes.'''

# ------------------------------------------------------------------------------
import types, copy
import appy.gen as gen
import po
from model import ModelClass
from utils import produceNiceMessage, getClassName
TABS = 4 # Number of blanks in a Python indentation

# ------------------------------------------------------------------------------
class Descriptor: # Abstract
    def __init__(self, klass, orderedAttributes, generator):
        # The corresponding Python class
        self.klass = klass
        # The names of the static appy-compliant attributes declared in
        # self.klass
        self.orderedAttributes = orderedAttributes
        # A reference to the code generator.
        self.generator = generator

    def __repr__(self): return '<Class %s>' % self.klass.__name__

class ClassDescriptor(Descriptor):
    '''This class gives information about an Appy class.'''

    def __init__(self, klass, orderedAttributes, generator):
        Descriptor.__init__(self, klass, orderedAttributes, generator)
        self.methods = '' # Needed method definitions will be generated here
        self.name = getClassName(self.klass, generator.applicationName)
        self.predefined = False
        self.customized = False
        # Phase and page names will be calculated later, when first required
        self.phases = None
        self.pages = None

    def getOrderedAppyAttributes(self, condition=None):
        '''Returns the appy types for all attributes of this class and parent
           class(es). If a p_condition is specified, ony Appy types matching
           the condition will be returned. p_condition must be a string
           containing an expression that will be evaluated with, in its context,
           "self" being this ClassDescriptor and "attrValue" being the current
           Type instance.
           
           Order of returned attributes already takes into account type's
           "move" attributes.'''
        attrs = []
        # First, get the attributes for the current class
        for attrName in self.orderedAttributes:
            try:
                attrValue = getattr(self.klass, attrName)
                hookClass = self.klass
            except AttributeError:
                attrValue = getattr(self.modelClass, attrName)
                hookClass = self.modelClass
            if isinstance(attrValue, gen.Field):
                if not condition or eval(condition):
                    attrs.append( (attrName, attrValue, hookClass) )
        # Then, add attributes from parent classes
        for baseClass in self.klass.__bases__:
            # Find the classDescr that corresponds to baseClass
            baseClassDescr = None
            for classDescr in self.generator.classes:
                if classDescr.klass == baseClass:
                    baseClassDescr = classDescr
                    break
            if baseClassDescr:
                attrs = baseClassDescr.getOrderedAppyAttributes() + attrs
        # Modify attributes order by using "move" attributes
        res = []
        for name, appyType, klass in attrs:
            if appyType.move:
                newPosition = len(res) - abs(appyType.move)
                if newPosition <= 0:
                    newPosition = 0
                res.insert(newPosition, (name, appyType, klass))
            else:
                res.append((name, appyType, klass))
        return res

    def getChildren(self):
        '''Returns, among p_allClasses, the classes that inherit from p_self.'''
        res = []
        for classDescr in self.generator.classes:
            if (classDescr.klass != self.klass) and \
               issubclass(classDescr.klass, self.klass):
                res.append(classDescr)
        return res

    def getPhases(self):
        '''Lazy-gets the phases defined on fields of this class.'''
        if not hasattr(self, 'phases') or (self.phases == None):
            self.phases = []
            for fieldName, appyType, klass in self.getOrderedAppyAttributes():
                if appyType.page.phase in self.phases: continue
                self.phases.append(appyType.page.phase)
        return self.phases

    def getPages(self):
        '''Lazy-gets the page names defined on fields of this class.'''
        if not hasattr(self, 'pages') or (self.pages == None):
            self.pages = []
            for fieldName, appyType, klass in self.getOrderedAppyAttributes():
                if appyType.page.name in self.pages: continue
                self.pages.append(appyType.page.name)
        return self.pages

    def getParents(self, allClasses):
        parentWrapper = 'AbstractWrapper'
        parentClass = '%s.%s' % (self.klass.__module__, self.klass.__name__)
        if self.klass.__bases__:
            baseClassName = self.klass.__bases__[0].__name__
            for k in allClasses:
                if self.klass.__name__ == baseClassName:
                    parentWrapper = '%s_Wrapper' % k.name
        return (parentWrapper, parentClass)

    def generateSchema(self):
        '''Generates i18n and other related stuff for this class.'''
        for attrName in self.orderedAttributes:
            try:
                attrValue = getattr(self.klass, attrName)
            except AttributeError:
                attrValue = getattr(self.modelClass, attrName)
            if not isinstance(attrValue, gen.Field): continue
            FieldDescriptor(attrName, attrValue, self).generate()

    def isAbstract(self):
        '''Is self.klass abstract?'''
        res = False
        if self.klass.__dict__.has_key('abstract'):
            res = self.klass.__dict__['abstract']
        return res

    def isRoot(self):
        '''Is self.klass root? A root class represents some kind of major
           concept into the application. For example, creating instances
           of such classes will be easy from the user interface.'''
        res = False
        if self.klass.__dict__.has_key('root'):
            res = self.klass.__dict__['root']
        return res

    def isFolder(self, klass=None):
        '''Must self.klass be a folder? If klass is not None, this method tests
           it on p_klass instead of self.klass.'''
        res = False
        theClass = self.klass
        if klass:
            theClass = klass
        if theClass.__dict__.has_key('folder'):
            res = theClass.__dict__['folder']
        else:
            if theClass.__bases__:
                res = self.isFolder(theClass.__bases__[0])
        return res

    def getCreators(self):
        '''Gets the specific creators defined for this class, excepted if
           attribute "creators" does not contain a list or roles.'''
        res = []
        if not hasattr(self.klass, 'creators'): return res
        if not isinstance(self.klass.creators, list): return res
        for creator in self.klass.creators:
            if isinstance(creator, gen.Role):
                if creator.local:
                    raise Exception('Local role "%s" cannot be used as a ' \
                                    'creator.' % creator.name)
                res.append(creator)
            else:
                res.append(gen.Role(creator))
        return res

    @staticmethod
    def getSearches(klass, tool=None):
        '''Returns the list of searches that are defined on this class. If
           p_tool is given, we are at execution time (not a generation time),
           and we may potentially execute search.show methods that allow to
           conditionnaly include a search or not.'''
        if klass.__dict__.has_key('search'):
            searches = klass.__dict__['search']
            if not tool: return searches
            # Evaluate attributes "show" for every search
            return [s for s in searches if s.isShowable(klass, tool)]
        return []

    @staticmethod
    def getSearch(klass, searchName):
        '''Gets the search named p_searchName.'''
        for search in ClassDescriptor.getSearches(klass):
            if search.name == searchName:
                return search

    def addIndexMethod(self, field, secondary=False):
        '''For indexed p_field, this method generates a method that allows to
           get the value of the field as must be copied into the corresponding
           index. Some fields have a secondary index for sorting purposes. If
           p_secondary is True, this method generates the method for this
           secondary index.'''
        m = self.methods
        spaces = TABS
        n = field.fieldName
        suffix = secondary and '_sort' or ''
        m += '\n' + ' '*spaces + 'def get%s%s%s(self):\n' % \
                                 (n[0].upper(), n[1:], suffix)
        spaces += TABS
        m += ' '*spaces + "'''Gets indexable value of field \"%s\".'''\n" % n
        suffix = secondary and ', True' or ''
        m += ' '*spaces + 'return self.getAppyType("%s").getIndexValue(' \
             'self%s)\n' % (n, suffix)
        self.methods = m
        if not secondary and field.appyType.hasSortIndex():
            self.addIndexMethod(field, secondary=True)

    def addField(self, fieldName, fieldType):
        '''Adds a new field to the Tool.'''
        exec "self.modelClass.%s = fieldType" % fieldName
        if fieldName in self.modelClass._appy_attributes:
            print('Warning, field "%s" is already existing on class "%s"' % \
                  (fieldName, self.modelClass.__name__))
            return
        self.modelClass._appy_attributes.append(fieldName)
        self.orderedAttributes.append(fieldName)

# ------------------------------------------------------------------------------
class WorkflowDescriptor(Descriptor):
    '''This class gives information about an Appy workflow.'''
    @staticmethod
    def getWorkflowName(klass):
        '''Returns the name of this workflow.'''
        res = klass.__module__.replace('.', '_') + '_' + klass.__name__
        return res.lower()

# ------------------------------------------------------------------------------
class FieldDescriptor:
    '''This class gathers information about a specific typed attribute defined
       in a gen-class.'''

    def __init__(self, fieldName, appyType, classDescriptor):
        self.appyType = appyType
        self.classDescr = classDescriptor
        self.generator = classDescriptor.generator
        self.applicationName = classDescriptor.generator.applicationName
        self.fieldName = fieldName
        self.fieldParams = {'name': fieldName}
        self.widgetParams = {}
        self.fieldType = None
        self.widgetType = None

    def i18n(self, id, default, nice=True):
        '''Shorthand for adding a new message into self.generator.labels'''
        self.generator.labels.append(id, default, nice=nice)

    def __repr__(self):
        return '<Field %s, %s>' % (self.fieldName, self.classDescr)

    def produceMessage(self, msgId, isLabel=True):
        '''Returns, as a tuple, parameters required for generating a i18n label
           via m_i18n.'''
        if isLabel:
            # Produce a nice default value based on the field name
            default = msgId.rsplit('_', 1)[-1]
            niceDefault = True
        else:
            # Produce a blank default value (for a "descr" or "help" label),
            # making it invisible by default.
            default = ' '
            niceDefault = False
        return msgId, default, niceDefault

    def mustGenerateLabels(self, field):
        '''Must we generate i18n labels for p_field ?

           Generating i18n labels for p_field (=p_self.appyType or some inner
           field) must not be done if it uses another field's labels or if it
           uses custom translations.
        '''
        return not field.label and not field.translations

    def generateLabels(self, field, prefix, forceLabel=False):
        '''Generate i18n labels for p_field. All labels will have a common
           p_prefix, that includes the class name and, potentially, the name of
           some outer field if p_field is an inner field.'''
        if forceLabel or field.hasLabel or field.generateLabel:
            self.i18n(*self.produceMessage(prefix))
        if field.hasDescr:
            descrId = prefix + '_descr'
            self.i18n(*self.produceMessage(descrId, isLabel=False))
        if field.hasHelp:
            helpId = prefix + '_help'
            self.i18n(*self.produceMessage(helpId, isLabel=False))

    def generateSelectLabels(self):
        '''Generate, when relevant, labels for a "select" field'''
        if type(self.appyType.validator) not in (list, tuple): return
        # Generate i18n messages for every possible value if the list
        # of values is fixed.
        for value in self.appyType.validator:
            label = '%s_%s_list_%s' % (self.classDescr.name,
                                       self.fieldName, value)
            self.i18n(label, value)

    def walkString(self):
        '''Generate String-specific i18n labels'''
        if self.appyType.isSelect: self.generateSelectLabels()

    def walkSelect(self):
        '''Generate Select-specific i18n labels'''
        self.generateSelectLabels()

    def walkBoolean(self):
        '''Generate Boolean-specific i18n labels'''
        if self.appyType.render == 'radios':
            for v in ('true', 'false'):
                label = '%s_%s_%s' % (self.classDescr.name, self.fieldName, v)
                self.i18n(label, self.appyType.yesNo[v])

    def walkAction(self):
        '''Generate Action-specific i18n labels'''
        if self.appyType.confirm:
            label = '%s_%s_confirm' % (self.classDescr.name, self.fieldName)
            self.i18n(label, po.CONFIRM, nice=False)
    walkPod = walkAction # Same i18n labels for Pod field

    def walkRef(self):
        '''Generate Ref-specific i18n labels'''
        # Add the label for the confirm message if relevant
        if self.appyType.addConfirm:
            label = '%s_%s_addConfirm' % (self.classDescr.name, self.fieldName)
            self.i18n(label, po.CONFIRM, nice=False)

    def walkList(self):
        '''Generate List-specific i18n labels'''
        for name, field in self.appyType.fields:
            if not self.mustGenerateLabels(field): continue
            prefix = '%s_%s_%s' % (self.classDescr.name, self.fieldName, name)
            self.generateLabels(field, prefix, forceLabel=True)
    walkDict = walkList # Same i18n labels for Dict field

    def walkCalendar(self):
        '''Generate Calendar-specific i18n labels'''
        eTypes = self.appyType.eventTypes
        if not isinstance(eTypes, list) and not isinstance(eTypes, tuple):return
        for et in self.appyType.eventTypes:
            label = '%s_%s_event_%s' % (self.classDescr.name,self.fieldName,et)
            self.i18n(label, et)

    def walkSwitch(self):
        '''Generate Calendar-specific i18n labels'''
        for fieldset, fields in self.appyType.fields:
            for name, field in fields:
                if not self.mustGenerateLabels(field): continue
                prefix = '%s_%s' % (self.classDescr.name, name)
                self.generateLabels(field, prefix, forceLabel=True)

    def walkField(self):
        '''Walks into the field definition and gathers data about the i18n
           labels.'''
        # Manage things common to all fields
        field = self.appyType
        # Put an index on this field ?
        if field.indexed and (self.fieldName != 'title'):
            self.classDescr.addIndexMethod(self)
        # Generate i18n labels for this field, when appropriate
        mustGenerateLabels = self.mustGenerateLabels(field)
        if mustGenerateLabels:
            prefix = '%s_%s' % (self.classDescr.name, self.fieldName)
            self.generateLabels(field, prefix)
        # Create i18n messages linked to pages and phases, only if there is more
        # than one page/phase for the class.
        if len(self.classDescr.getPhases()) > 1:
            # Create the message for the name of the phase
            phaseName = field.page.phase
            msgId = '%s_phase_%s' % (self.classDescr.name, phaseName)
            self.i18n(msgId, phaseName)
        if (len(self.classDescr.getPages()) > 1) and not field.page.label:
            # Create the message for the name of the page
            pageName = field.page.name
            # Avoid to generate labels for main pages of predefined classes
            if not ((pageName == 'main') and self.classDescr.predefined):
                msgId = '%s_page_%s' % (self.classDescr.name, pageName)
                self.i18n(msgId, pageName)
        # Create i18n messages linked to groups
        for group in field.getGroups():
            if group.label: continue
            group.generateLabels(self.generator.labels, self.classDescr, set())
        # Generate type-specific i18n labels
        if mustGenerateLabels:
            method = 'walk%s' % field.type
            if hasattr(self, method): getattr(self, method)()

    def generate(self):
        '''Generates the i18n labels for this field'''
        self.walkField()

# ------------------------------------------------------------------------------
class ToolClassDescriptor(ClassDescriptor):
    '''Represents the POD-specific fields that must be added to the tool.'''
    def __init__(self, klass, generator):
        ClassDescriptor.__init__(self,klass,klass._appy_attributes[:],generator)
        self.modelClass = self.klass
        self.predefined = True
        self.customized = False

    def getParents(self, allClasses=()):
        res = ['Tool']
        if self.customized:
            res.append('%s.%s' % (self.klass.__module__, self.klass.__name__))
        return res

    def update(self, klass, attributes):
        '''This method is called by the generator when he finds a custom tool
           definition. We must then add the custom tool elements in this default
           Tool descriptor.'''
        self.orderedAttributes += attributes
        self.klass = klass
        self.customized = True

    def isFolder(self, klass=None): return True
    def isRoot(self): return False

class UserClassDescriptor(ClassDescriptor):
    '''Appy-specific class for representing a user.'''

    def __init__(self, klass, generator):
        ClassDescriptor.__init__(self,klass,klass._appy_attributes[:],generator)
        self.modelClass = self.klass
        self.predefined = True
        self.customized = False

    def getParents(self, allClasses=()):
        res = ['User']
        if self.customized:
            res.append('%s.%s' % (self.klass.__module__, self.klass.__name__))
        return res

    def update(self, klass, attributes):
        '''This method is called by the generator when he finds a custom user
           definition. We must then add the custom user elements in this
           default User descriptor.'''
        self.orderedAttributes += attributes
        self.klass = klass
        self.customized = True

    def isFolder(self, klass=None): return True

class GroupClassDescriptor(ClassDescriptor):
    '''Represents the class that corresponds to the Group for the generated
       application.'''

    def __init__(self, klass, generator):
        ClassDescriptor.__init__(self,klass,klass._appy_attributes[:],generator)
        self.modelClass = self.klass
        self.predefined = True
        self.customized = False

    def getParents(self, allClasses=()):
        res = ['Group']
        if self.customized:
            res.append('%s.%s' % (self.klass.__module__, self.klass.__name__))
        return res

    def update(self, klass, attributes):
        '''This method is called by the generator when he finds a custom group
           definition. We must then add the custom group elements in this
           default Group descriptor.

           NOTE: currently, it is not possible to define a custom Group
           class.'''
        self.orderedAttributes += attributes
        self.klass = klass
        self.customized = True

    def isFolder(self, klass=None): return False

class TranslationClassDescriptor(ClassDescriptor):
    '''Represents the set of translation ids for a gen-application.'''

    def __init__(self, klass, generator):
        ClassDescriptor.__init__(self,klass,klass._appy_attributes[:],generator)
        self.modelClass = self.klass
        self.predefined = True
        self.customized = False

    def getParents(self, allClasses=()): return ('Translation',)
    def isFolder(self, klass=None): return False

    def addLabelField(self, messageId, page):
        '''Adds a Computed field that will display, in the source language, the
           content of the text to translate.'''
        field = gen.Computed(method=self.modelClass.label, plainText=False,
                             page=page, show=self.modelClass.show, layouts='f')
        self.addField('%s_label' % messageId, field)

    def addMessageField(self, messageId, page, i18nFiles):
        '''Adds a message field corresponding to p_messageId to the Translation
           class, on a given p_page. We need i18n files p_i18nFiles for
           fine-tuning the String type to generate for this field (one-line?
           several lines?...)'''
        params = {'page':page, 'layouts':'f', 'show': self.modelClass.show}
        appName = self.generator.applicationName
        # Scan all messages corresponding to p_messageId from all translation
        # files. We will define field length from the longer found message
        # content.
        maxLine = 100 # We suppose a line is 100 characters long.
        width = 0
        height = 0
        for fileName, poFile in i18nFiles.iteritems():
            if not fileName.startswith('%s-' % appName) or \
               not i18nFiles[fileName].messagesDict.has_key(messageId):
                # In this case this is not one of our Appy-managed translation
                # files.
                continue
            msgContent = i18nFiles[fileName].messagesDict[messageId].msg
            # Compute width
            width = max(width, len(msgContent))
            # Compute height (a "\n" counts for one line)
            mHeight = int(len(msgContent)/maxLine) + msgContent.count('<br/>')
            height = max(height, mHeight)
        if height < 1:
            # This is a one-line field
            params['width'] = width
        else:
            # This is a multi-line field, or a very-long-single-lined field
            params['format'] = gen.String.TEXT
            params['height'] = height
        self.addField(messageId, gen.String(**params))

class PageClassDescriptor(ClassDescriptor):
    '''Represents the class that corresponds to a Page.'''

    def __init__(self, klass, generator):
        ClassDescriptor.__init__(self,klass,klass._appy_attributes[:],generator)
        self.modelClass = self.klass
        self.predefined = True
        self.customized = False

    def getParents(self, allClasses=()):
        res = ['Page']
        if self.customized:
            res.append('%s.%s' % (self.klass.__module__, self.klass.__name__))
        return res

    def update(self, klass, attributes):
        '''This method is called by the generator when he finds a custom page
           definition. We must then add the custom page elements in this
           default Page descriptor.

           NOTE: currently, it is not possible to define a custom Page class.'''
        self.orderedAttributes += attributes
        self.klass = klass
        self.customized = True

    def isFolder(self, klass=None): return True
# ------------------------------------------------------------------------------



'''Module managing database locks set on object pages'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from DateTime import DateTime
from persistent.mapping import PersistentMapping

from appy.px import Px

# Errors - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
LOCKED_OTHER = 'This page was locked by someone else.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Lock:
    '''A lock is a database structure remembering which user (login) has locked
      which page on which object. Locks on a given database object is
      implemented as a dict object.locks of the form
                  ~{s_page: (s_userId, DateTime_lockDate)}~
    '''

    @classmethod
    def set(class_, o, user, page=None, field=None):
        '''A p_user edits a given p_page on p_o: we will set a lock, to prevent
           other users to edit this page at the same time.'''
        # When setting a lock from a field being inline-edited, p_page is not
        # given and p_field is given instead. In that case, the page will be
        # retrieved from the field, but the lock will not be set if the field is
        # not persistent or whose persistence is managed by a specific method.
        if field and (not field.persist or callable(field.persist)): return
        # This operation requires a database commit
        o.H().commit = True
        # If it does not exist yet, create, on p_o, the persistent mapping that
        # will store the lock.
        if not hasattr(o, 'locks'):
            # ~{s_page: (s_userLogin, DateTime_lockDate)}~
            o.locks = PersistentMapping()
        # Raise an error is the page is already locked by someone else. If the
        # page is already locked by the same user, we don't mind: he could have
        # used its browser's back / forward buttons...
        login = user.login
        locks = o.locks
        page = page or field.page.name
        if (page in locks) and (login != locks[page][0]):
            login, date = locks[page]
            locker = o.search1('User', secure=False, login=login)
            map = {'user': locker.getTitle(), 'date': o.tool.formatDate(date)}
            o.raiseUnauthorized(o.translate('page_locked', mapping=map))
        # Set the lock
        locks[page] = (login, DateTime())

    @classmethod
    def isSet(class_, o, user, page):
        '''Is p_page locked on p_o? If the page is locked by the p_user, we
           don't mind and consider the page as unlocked. If the page is locked,
           this method returns the tuple (login, lockDate).'''
        if hasattr(o, 'locks') and (page in o.locks):
            if (user.login != o.locks[page][0]): return o.locks[page]

    @classmethod
    def remove(class_, o, page=None, field=None, force=False):
        '''Removes the lock on p_o's p_page. This happens:
           - after the page has been saved: the lock must be released;
           - when an admin wants to force the deletion of a lock that was left
             on p_page for too long (p_force=True).
        '''
        # When removing a lock from a field being inline-edited, p_page is not
        # given and p_field is given instead.
        page = page or field.page.name
        locks = getattr(o, 'locks', None)
        if not locks or (page not in locks): return
        # If there is a custom persistence for p_field, the lock was not set
        if field and (not field.persist or callable(field.persist)): return
        # Raise an error if the user saving changes is not the one that has
        # locked the page (excepted if p_force is True).
        if not force:
            login = o.user.login
            if locks[page][0] != login:
                o.raiseUnauthorized(LOCKED_OTHER)
        # Remove the lock
        del(locks[page])
        # This requires a database commit
        o.H().commit = True

    @classmethod
    def removeMine(class_, o, user, page):
        '''If p_user has set a lock on p_page, this method removes it. This
           method is called when the user that locked a page consults PX "view"
           for this page. In this case, we consider that the user has left the
           edit page in an unexpected way and we remove the lock.'''
        if hasattr(o, 'locks') and (page in o.locks) and \
           (user.login == o.locks[page][0]):
            del(o.locks[page])

    @classmethod
    def getLockers(class_, o):
        '''Return the list of logins having locked at least one page on p_o'''
        r = set()
        if hasattr(o, 'locks'):
            for login, date in o.locks.values():
                r.add(login)
        return r

    @classmethod
    def unlockableBy(class_, o, user):
        '''May the currently logged p_user unlock p_o ?'''
        for role in o.config.database.unlockers:
            if isinstance(role, str):
                condition = user.hasRole(role)
            else:
                if role.local:
                    condition = user.hasRole(role.name, o)
                else:
                    condition = user.hasRole(role.name)
            if condition:
                return True

    @classmethod
    def unlock(class_, handler):
        '''Called when the logged user wants to remove a lock that was left for
           too long by some user.'''
        req = handler.req
        o = handler.tool.getObject(req.objectId)
        Lock.remove(o, req.page, force=True)
        handler.resp.goto(message=o.translate('action_done'))

    # Shows icons instead of the "edit" icon/button when a lock is defined
    px = Px('''
     <a if="editable and locked">
      <img class=":'help ' + lockStyle"
       var="unlock=o.Lock.unlockableBy(o, user);
            lockDate=tool.Date.format(tool, locked[1]);
            lockMap={'user':user.getTitleFromLogin(locked[0]),'date':lockDate};
            lockMsg=_('page_locked', mapping=lockMap);
            lockTxt='%s %s' % (lockMsg,_('page_unlock')) if unlock else lockMsg"
       src=":url('locked.svg')" title=":lockTxt"
       onclick=":'onUnlockPage(%s,%s)' % (q(o.url),q(page)) if unlock else ''"/>
     </a>''')
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
from DateTime import DateTime
from cStringIO import StringIO
import time, os.path, mimetypes, shutil, base64

from appy.px import Px
from appy import Object
from appy.fields import Field, Layouts
from appy.shared import utils as sutils
from appy.shared import UnmarshalledFile, mimeTypesExts

# ------------------------------------------------------------------------------
WRONG_FILE_TUPLE = 'This is not the way to set a file. You can specify a ' \
    '2-tuple (fileName, fileContent) or a 3-tuple (fileName, fileContent, ' \
    'mimeType).'
CONVERSION_ERROR = 'An error occurred. %s'
RESIZED = '%s resized to %spx.'

def guessMimeType(fileName):
    '''Try to find the MIME type of file p_fileName'''
    return mimetypes.guess_type(fileName)[0] or File.defaultMimeType

def osPathJoin(*pathElems):
    '''Version of os.path.elems that takes care of path elems being empty
       strings.'''
    return os.path.join(*pathElems).rstrip(os.sep)

def getShownSize(size):
    '''Express p_size (a file size in bytes) in a human-readable way'''
    # Display the size in bytes if smaller than 1024 bytes
    if size < 1024: return '%d byte(s)' % size
    size = size / 1024.0 # This is the size, in Kb
    if size < 1024: return '%s Kb' % sutils.formatNumber(size, precision=1)
    size = size / 1024.0 # This is the size, in Mb
    return '%s Mb' % sutils.formatNumber(size, precision=1)

# ------------------------------------------------------------------------------
class FileInfo:
    '''A FileInfo instance holds metadata about a file on the filesystem.

       For every File field, we will store a FileInfo instance in the dabatase;
       the real file will be stored in the Appy/ZODB database-managed
       filesystem.

       This is the primary usage of FileInfo instances. FileInfo instances can
       also be used every time we need to manipulate a file. For example, when
       getting the content of a Pod field, a temporary file may be generated and
       you will get a FileInfo that represents it.
    '''
    BYTES = 50000
    NOT_FOUND = 'File "%s" was not found.'

    def __init__(self, fsPath, inDb=True, uploadName=None):
        '''p_fsPath is the path of the file on disk.
           - If p_inDb is True, this FileInfo will be stored in the database and
             will hold metadata about a File field whose content will lie in the
             database-controlled filesystem. In this case, p_fsPath is the path
             of the file *relative* to the root DB folder. We avoid storing
             absolute paths in order to ease the transfer of databases from one
             place to the other. Moreover, p_fsPath does not include the
             filename, that will be computed later, from the field name.

           - If p_inDb is False, this FileInfo is a simple temporary object
             representing any file on the filesystem (not necessarily in the
             db-controlled filesystem). For instance, it could represent a temp
             file generated from a Pod field in the OS temp folder. In this
             case, p_fsPath is the absolute path to the file, including the
             filename. If you manipulate such a FileInfo instance, please avoid
             using methods that are used by Appy to manipulate
             database-controlled files (like methods getFilePath, removeFile,
             writeFile or copyFile).'''
        self.fsPath = fsPath
        self.fsName = None # The name of the file in fsPath
        self.uploadName = uploadName # The name of the uploaded file
        self.size = 0 # Its size, in bytes
        self.mimeType = None # Its MIME type
        self.modified = None # The last modification date for this file
        # Complete metadata if p_inDb is False. p_inDb is not stored as is:
        # checking if self.fsName is the empty string is equivalent.
        if not inDb:
            self.fsName = '' # Already included in self.fsPath
            self.setAttributes(self.fsPath)

    def setAttributes(self, path):
        '''Compute file attributes (size, MIME type and last modification date)
           for a file whose absolute p_path is passed and store these attributes
           on p_self.'''
        info = os.stat(path)
        self.size = info.st_size
        self.mimeType = guessMimeType(path)
        self.modified = DateTime(info.st_mtime)

    def getFilePath(self, obj):
        '''Returns the absolute file name of the file on disk that corresponds
           to this FileInfo instance.'''
        dbFolder, folder = obj.o.getFsFolder()
        r = osPathJoin(dbFolder, folder, self.fsName)
        if not os.path.exists(r):
            # It may already have been deleted by a failed transaction. Try to
            # get a copy we may have made in the OS temp folder.
            folder = os.path.basename(folder)
            r = osPathJoin(sutils.getOsTempFolder(), folder, self.fsName)
        return r

    def removeFile(self, dbFolder='', removeEmptyFolders=False):
        '''Removes the file from the filesystem.'''
        try:
            os.remove(osPathJoin(dbFolder, self.fsPath, self.fsName))
        except Exception, e:
            # If the current ZODB transaction is re-triggered, the file may
            # already have been deleted.
            pass
        # Don't leave empty folders on disk. So delete folder and parent folders
        # if this removal leaves them empty (unless p_removeEmptyFolders is
        # False).
        if removeEmptyFolders:
            sutils.FolderDeleter.deleteEmpty(osPathJoin(dbFolder,self.fsPath))

    def normalizeFileName(self, name):
        '''Normalizes file p_name.'''
        return name[max(name.rfind('/'), name.rfind('\\'), name.rfind(':'))+1:]

    def getShownSize(self): return getShownSize(self.size)

    def replicateFile(self, src, dest):
        '''p_src and p_dest are open file handlers. This method copies content
           of p_src to p_dest and returns the file size. Note that p_src can
           also be binary data in a string.'''
        size = 0
        if isinstance(src, str): src = StringIO(src)
        while True:
            chunk = src.read(self.BYTES)
            if not chunk: break
            size += len(chunk)
            dest.write(chunk)
        return size

    def getBase64(self, obj=None):
        '''Returns the file content, as a base64-encoded string'''
        if obj:
            filePath = self.getFilePath(obj)
        else:
            filePath = self.fsPath
        f = file(filePath, 'rb')
        res = base64.b64encode(f.read())
        f.close()
        return res

    def getExtension(self):
        '''Get the file extension from the MIME type or from the upload name'''
        if self.mimeType in mimeTypesExts:
            return mimeTypesExts[self.mimeType]
        elif self.uploadName:
            parts = os.path.splitext(self.uploadName)
            if len(parts) > 1:
                return parts[-1][1:]

    def getMimeTypeFromFileUpload(self, fileObj):
        '''Under some unknown circumstances, the MIME type received from Zope
           FileUpload instances is wrong:
           * MIME type of docx and xlsx documents may be wrongly initialised to
             "application/zip";
           * MIME type of some Excel (.xls) files have MIME type
             "application/msword".
           This method corrects it.'''
        mimeType = fileObj.headers.get('content-type')
        ext = os.path.splitext(fileObj.filename)[1]
        # If no extension is there, I cannot correct the MIME type
        if not ext: return mimeType
        # Correct xls files having MIME type "application/msword"
        if (ext == '.xls') and (mimeType == 'application/msword'):
            return 'application/vnd.ms-excel'
        # No error: return the MIME type as computed by Zope
        if not ext or (mimeType != 'application/zip') or (ext == '.zip'):
            return mimeType
        # Correct the wrong MIME type
        ext = ext[1:].lower()
        for mime, extension in mimeTypesExts.iteritems():
            if extension == ext: return mime
        # If we are here, we haven't found the correct MIME type
        return mimeType

    def writeFile(self, fieldName, fileObj, dbFolder):
        '''Writes to the filesystem the p_fileObj file, that can be:
           - a Zope FileUpload (coming from a HTTP post);
           - a OFS.Image.File object (legacy within-ZODB file object);
           - another ("not-in-DB") FileInfo instance;
           - a tuple (fileName, fileContent, mimeType)
             (see doc in method File.store below).'''
        # Determine p_fileObj's type
        fileType = fileObj.__class__.__name__
        # Determine the MIME type and the base name of the file to store
        if fileType == 'FileUpload':
            mimeType = self.getMimeTypeFromFileUpload(fileObj)
            fileName = fileObj.filename
        elif fileType == 'File':
            mimeType = fileObj.content_type
            fileName = fileObj.filename
        elif fileType == 'FileInfo':
            mimeType = fileObj.mimeType
            fileName = fileObj.uploadName
        else:
            mimeType = fileObj[2]
            fileName = fileObj[0]
        self.mimeType = mimeType or File.defaultMimeType
        if not fileName:
            # Name it according to field name. Deduce file extension from the
            # MIME type.
            ext = (self.mimeType in mimeTypesExts) and \
                  mimeTypesExts[self.mimeType] or 'bin'
            fileName = '%s.%s' % (fieldName, ext)
        # As a preamble, extract file metadata from p_fileObj and store it in
        # this FileInfo instance.
        name = self.normalizeFileName(fileName)
        self.uploadName = name
        self.fsName = '%s%s' % (fieldName, os.path.splitext(name)[1].lower())
        # Write the file on disk (and compute/get its size in bytes)
        fsName = osPathJoin(dbFolder, self.fsPath, self.fsName)
        f = file(fsName, 'wb')
        if fileType == 'FileUpload':
            # Write the FileUpload instance on disk
            self.size = self.replicateFile(fileObj, f)
        elif fileType == 'File':
            # Write the File instance on disk
            if fileObj.data.__class__.__name__ == 'Pdata':
                # The file content is splitted in several chunks
                f.write(fileObj.data.data)
                nextPart = fileObj.data.next
                while nextPart:
                    f.write(nextPart.data)
                    nextPart = nextPart.next
            else:
                # Only one chunk
                f.write(fileObj.data)
            self.size = fileObj.size
        elif fileType == 'FileInfo':
            src = file(fileObj.fsPath, 'rb')
            self.size = self.replicateFile(src, f)
            src.close()
        else:
            # Write fileObj[1] on disk
            if fileObj[1].__class__.__name__ == 'file':
                # It is an open file handler
                self.size = self.replicateFile(fileObj[1], f)
            else:
                # We have file content directly in fileObj[1]
                self.size = len(fileObj[1])
                f.write(fileObj[1])
        f.close()
        from DateTime import DateTime
        self.modified = DateTime()

    def copyFile(self, fieldName, filePath, dbFolder):
        '''Copies the "external" file stored at p_filePath in the db-controlled
           file system, for storing a value for p_fieldName.'''
        # Set names for the file
        name = self.normalizeFileName(filePath)
        self.uploadName = name
        self.fsName = '%s%s' % (fieldName, os.path.splitext(name)[1])
        # Set mimeType
        self.mimeType = guessMimeType(filePath)
        # Copy the file
        fsName = osPathJoin(dbFolder, self.fsPath, self.fsName)
        shutil.copyfile(filePath, fsName)
        from DateTime import DateTime
        self.modified = DateTime()
        self.size = os.stat(fsName).st_size

    def writeResponse(self, response, dbFolder='',
                      disposition='attachment', enableCache=False):
        '''Writes this file in the HTTP p_response object.

           For privacy reasons, p_enableCache is disabled by default. This way,
           it cannot be stored in the browser cache. For non-privacy-sensitive
           information (ie, icons), you can enable caching. When enabled, the
           last modification date is returned in the HTTP header: this way, the
           browser will only download it if the cached version is older than the
           last version, according to this date.
        '''
        set = response.setHeader
        # The file may not exist on disk
        fsName = osPathJoin(dbFolder, self.fsPath, self.fsName)
        if not os.path.isfile(fsName):
            # Return a dummy file containing an error message
            msg = self.NOT_FOUND % sutils.normalizeString(self.uploadName)
            set('Content-Type', 'text/plain')
            set('Content-Length', len(msg))
            response.write(msg)
            return
        # Initialise response headers
        set('Content-Type', self.mimeType)
        set('Content-Length', self.size)
        set('Content-Disposition',
            '%s;filename="%s"' % (disposition, self.uploadName))
        # For now, disable byte serving (value "bytes" instead of "none")
        set('Accept-Ranges', 'none')
        if enableCache:
            set('Last-Modified', self.modified.rfc822())
        else:
            set('Cache-Control', 'no-cache, no-store, must-revalidate')
            set('Expires', '0')
        # Write the file in the response
        f = file(fsName, 'rb')
        while True:
            chunk = f.read(self.BYTES)
            if not chunk: break
            response.write(chunk)
        f.close()

    def dump(self, obj, filePath=None, format=None):
        '''Exports this file to disk (outside the db-controlled filesystem).
           The tied Appy p_obj(ect) is required. If p_filePath is specified, it
           is the path name where the file will be dumped; folders mentioned in
           it must exist. If not, the file will be dumped in the OS temp folder.
           The absolute path name of the dumped file is returned. If an error
           occurs, the method returns None. If p_format is specified,
           LibreOffice will be called for converting the dumped file to the
           desired format.'''
        if not filePath:
            filePath = '%s/file%f.%s' % (sutils.getOsTempFolder(), time.time(),
                                         self.fsName)
        # Copies the file to disk
        shutil.copyfile(self.getFilePath(obj), filePath)
        if format:
            # Convert the dumped file using LibreOffice
            out, err = obj.tool.convert(filePath, format)
            # Even if we have an "error" message, it could be a simple warning.
            # So we will continue here and, as a subsequent check for knowing if
            # an error occurred or not, we will test the existence of the
            # converted file (see below).
            os.remove(filePath)
            # Get the name of the converted file
            baseName, ext = os.path.splitext(filePath)
            if (ext == '.%s' % format):
                filePath = '%s.res.%s' % (baseName, format)
            else:
                filePath = '%s.%s' % (baseName, format)
            if not os.path.exists(filePath):
                obj.log(CONVERSION_ERROR % err, type='error')
                return
        return filePath

    def resize(self, folder, width, obj):
        '''Resize this image if it is a resizable image'''
        if self.mimeType not in File.resizableImages: return
        # Get the absolute path to the file on disk
        path = self.getFilePath(obj)
        # Get the width, in pixels, that will be used to resize it
        width = sutils.keepDigits(str(width))
        sutils.convert(path, '-resize %sx%s>' % (width, width))
        obj.log(RESIZED % (path, width))
        # (re)compute p_self's attributes, after resizing
        self.setAttributes(path)

# ------------------------------------------------------------------------------
class File(Field):

    # MIME types for images
    imageMimeTypes = ('image/png', 'image/jpeg', 'image/gif')
    # MIME types for resizable images
    resizableImages = imageMimeTypes

    pxView = pxCell = Px('''
      <x>::field.getDownloadLink(name, zobj, layoutType)</x>''')

    pxEdit = Px('''
     <x var="fName='%s_file' % name">
      <x if="value">:field.pxView</x><br if="value"/>
      <x if="value">
       <!-- Keep the file unchanged -->
       <input type="radio" value="nochange"
              checked=":value and 'checked' or None"
              name=":'%s_delete' % name" id=":'%s_nochange' % name"
              onclick=":'document.getElementById(%s).disabled=true'% q(fName)"/>
       <label lfor=":'%s_nochange' % name">:_('keep_file')</label><br/>
       <!-- Delete the file -->
       <x if="not field.required">
        <input type="radio" value="delete"
               name=":'%s_delete' % name" id=":'%s_delete' % name"
               onclick=":'document.getElementById(%s).disabled=true'%q(fName)"/>
        <label lfor=":'%s_delete' % name">:_('delete_file')</label><br/>
       </x>
       <!-- Replace with a new file -->
       <input type="radio" value=""
              checked=":not value and 'checked' or None"
              name=":'%s_delete' % name" id=":'%s_upload' % name"
              onclick=":'document.getElementById(%s).disabled=false'%q(fName)"/>
       <label lfor=":'%s_upload' % name">:_('replace_file')</label><br/>
      </x>
      <!-- The upload field -->
      <input type="file" name=":fName" id=":fName" style=":field.getStyle()"
             onChange=":field.getJsOnChange()"/>
      <script var="isDisabled=not value and 'false' or 'true'"
             type="text/javascript">:'document.getElementById(%s).disabled=%s'%\
                                     (q(fName), isDisabled)</script></x>''')

    pxSearch = ''

    def __init__(self, validator=None, multiplicity=(0,1), default=None,
      defaultOnEdit=None, show=True, page='main', group=None, layouts=None,
      move=0, specificReadPermission=False, specificWritePermission=False,
      width=None, height=None, inputWidth=None, maxChars=None, colspan=1,
      master=None, masterValue=None, focus=False, historized=False,
      mapping=None, generateLabel=None, label=None, isImage=False,
      downloadAction=None, sdefault='', scolspan=1, swidth=None, sheight=None,
      view=None, cell=None, edit=None, xml=None, translations=None, render=None,
      icon='paperclip', nameStorer=None, resize=False):
        # This boolean is True if the file is an image
        self.isImage = isImage
        # "downloadAction" can be a method called every time the file is
        # downloaded. The method gets, as single arg, the FileInfo instance
        # representing the downloaded file.
        self.downloadAction = downloadAction
        # If "render" is "icon", the file will be rendered as an icon
        # (self.icon) on "buttons" and "result" layouts.
        self.render = render
        # Icon to use when this file is rendered as an icon
        self.icon = icon
        # In "nameStorer", you can specify another field that will store the
        # file name. This field must be a String belonging to the same class as
        # p_self. As soon as, in the UI, a file is selected in p_self's widget,
        # its name will be copied into the nameStorer field, without the
        # extension, only if this latter is not empty.
        self.nameStorer = nameStorer
        # Attribute "width" is used to specify the width of the image or
        # document. The width of the input field allowing to upload the file can
        # be defined in attribute "inputWidth".
        self.inputWidth = inputWidth
        # If attribute "resize" is False (the default), the image will be
        # previewed in the UI with dimensions as defined in attributes "width"
        # and "height", but without being effectively resized. If "resize" is
        # True, when uploading the file, it will be resized to self.width,
        # keeping the width / height ratio.
        self.resize = resize
        # Call the base constructor
        Field.__init__(self, validator, multiplicity, default, defaultOnEdit,
          show, page, group, layouts, move, False, True, None, False,
          specificReadPermission, specificWritePermission, width, height, None,
          colspan, master, masterValue, focus, historized, mapping,
          generateLabel, label, sdefault, scolspan, swidth, sheight, True,
          False, view, cell, edit, xml, translations)

    def getRequestValue(self, obj, requestName=None):
        name = requestName or self.name
        return obj.REQUEST.get('%s_file' % name)

    def getRequestSuffix(self): return '_file'

    def getCopyValue(self, obj):
        '''Create a copy of the FileInfo instance stored for p_obj for this
           field. This copy will contain the absolute path to the file on the
           filesystem. This way, the file may be read independently from p_obj
           (and copied somewhere else).'''
        info = self.getValue(obj)
        if not info: return
        # Create a "not-in-DB", temporary FileInfo
        return FileInfo(info.getFilePath(obj), inDb=False,
                        uploadName=info.uploadName)

    def getDefaultLayouts(self):
        '''Default layouts depend on the field being in a grid or not'''
        return self.inGrid() and Layouts.File.bg or Layouts.File.b

    def isEmptyValue(self, obj, value):
        '''Must p_value be considered as empty?'''
        if value: return
        # If "nochange", the value must not be considered as empty
        if hasattr(obj, 'REQUEST'):
            return obj.REQUEST.get('%s_delete' % self.name) != 'nochange'
        return True

    def getJsOnChange(self):
        '''Gets the JS code for updaing the name storer when defined'''
        storer = self.nameStorer
        if not storer: return ''
        return 'updateFileNameStorer(this, "%s")' % storer.name

    def getStyle(self):
        '''Get the content of the "style" attribute of the "input" tag on the
           "edit" layout for this field.'''
        if self.inputWidth:
            return 'width: %s' % self.inputWidth
        return ''

    def isRenderable(self, layoutType):
        '''A file with 'icon' rendering is potentially renderable everywhere'''
        if self.render == 'icon': return True
        return layoutType != 'buttons'

    def getDownloadLink(self, name, obj, layoutType):
        '''Gets the HTML code for downloading the file p_value as stored in
           field p_name on p_obj.'''
        value = self.getValueIf(obj, name, layoutType)
        # Display an empty value
        if not value: return (layoutType != 'cell') and '-' or ''
        # Get the "file title", derived from the upload name and size
        size = value.getShownSize()
        title = "%s - %s" % (value.uploadName, size)
        # On "edit", simply repeat the file title
        if layoutType == 'edit': return value.uploadName
        # Build the URL for downloading or displaying the file
        url = '%s/download?name=%s' % (obj.absolute_url(), name)
        # For images, display them directly
        if self.isImage:
            # Define a max width when relevant
            if self.width:
                css = ' style="max-width: %s"' % self.width
            else:
                css = ''
            return '<img src="%s" title="%s"%s/>' % (url, title, css)
        # For non-images, display a link for downloading it, as an icon when
        # relevant.
        if self.render == 'icon':
            content = '<img src="%s" title="%s"/>' % \
                      (obj.getTool().getIncludeUrl(self.icon), title)
            # On "view", we have place, so display "title" besides the icon
            suffix = (layoutType == 'view') and title or ''
        else:
            # Display textual information only
            content = title
            suffix = ''
        # Style the suffix
        if suffix: suffix = '<span class="refLink">%s</span>' % suffix
        return '<a href="%s">%s%s</a>' % (url, content, suffix)

    imageExts = ('.jpg', '.jpeg', '.png', '.gif')
    def validateValue(self, obj, value):
        form = obj.REQUEST.form
        action = '%s_delete' % self.name
        if (not value or not value.filename) and form.has_key(action) and \
            not form[action]:
            # If this key is present but empty, it means that the user selected
            # "replace the file with a new one". So in this case he must provide
            # a new file to upload.
            return obj.translate('file_required')
        # Check that, if self.isImage, the uploaded file is really an image
        if value and value.filename and self.isImage:
            ext = os.path.splitext(value.filename)[1].lower()
            if ext not in File.imageExts:
                return obj.translate('image_required')

    defaultMimeType = 'application/octet-stream'
    def store(self, obj, value):
        '''Stores the p_value that represents some file. p_value can be:
           a. an instance of Zope class ZPublisher.HTTPRequest.FileUpload. In
              this case, it is file content coming from a HTTP POST;
           b. an instance of Zope class OFS.Image.File (legacy within-ZODB file
              object);
           c. an instance of appy.shared.UnmarshalledFile. In this case, the
              file comes from a peer Appy site, unmarshalled from XML content
              sent via an HTTP request;
           d. a string. In this case, the string represents the path of a file
              on disk;
           e. a 2-tuple (fileName, fileContent) where:
              - fileName is the name of the file (ie "myFile.odt")
              - fileContent is the binary or textual content of the file or an
                open file handler.
           f. a 3-tuple (fileName, fileContent, mimeType) where
              - fileName and fileContent have the same meaning than above;
              - mimeType is the MIME type of the file.
           g. a FileInfo instance, that must be "not-in-DB", ie, with an
              absolute path in attribute fsPath.
        '''
        zobj = obj.o
        if value:
            # There is a new value to store. Get the folder on disk where to
            # store the new file.
            dbFolder, folder = zobj.getFsFolder(create=True)
            # Remove the previous file if it existed
            info = getattr(obj.aq_base, self.name, None)
            if info:
                # The previous file can be a legacy File object in an old
                # database we are migrating.
                if isinstance(info, FileInfo): info.removeFile(dbFolder)
                else: delattr(obj, self.name)
            # Store the new file. As a preamble, create a FileInfo instance.
            info = FileInfo(folder)
            cfg = zobj.getProductConfig()
            if isinstance(value, cfg.FileUpload) or isinstance(value, cfg.File):
                # Cases a, b
                value.filename = value.filename.replace('/', '-')
                info.writeFile(self.name, value, dbFolder)
            elif isinstance(value, UnmarshalledFile):
                # Case c
                fileInfo = (value.name, value.content, value.mimeType)
                info.writeFile(self.name, fileInfo, dbFolder)
            elif isinstance(value, basestring):
                # Case d
                info.copyFile(self.name, value, dbFolder)
            elif isinstance(value, FileInfo):
                # Case g
                info.writeFile(self.name, value, dbFolder)
            else:
                # Cases e, f. Extract file name, content and MIME type.
                fileName = mimeType = None
                if len(value) == 2:
                    fileName, fileContent = value
                elif len(value) == 3:
                    fileName, fileContent, mimeType = value
                if not fileName:
                    raise Exception(WRONG_FILE_TUPLE)
                mimeType = mimeType or guessMimeType(fileName)
                info.writeFile(self.name, (fileName, fileContent, mimeType),
                               dbFolder)
            # Resize the image when relevant
            if self.resize and self.width: info.resize(folder, self.width, obj)
            # Store the FileInfo instance in the database
            setattr(obj, self.name, info)
        else:
            # I store value "None", excepted if I find in the request the desire
            # to keep the file unchanged.
            action = None
            rq = getattr(zobj, 'REQUEST', None)
            if rq: action = rq.get('%s_delete' % self.name, None)
            if action != 'nochange':
                # Delete the file on disk
                info = getattr(zobj.aq_base, self.name, None)
                if info:
                    info.removeFile(zobj.getDbFolder(), removeEmptyFolders=True)
                # Delete the FileInfo in the DB
                setattr(zobj, self.name, None)

    def onDownload(self, obj, rq):
        '''Triggered when a file download is requested from the ui'''
        # Security check
        obj.mayView(self.readPermission, raiseError=True)
        # Write the file in the HTTP response
        info = getattr(obj.aq_base, self.name, None)
        if info:
            # Content disposition may be given in the request
            disposition = rq.get('disposition', 'attachment')
            if disposition not in ('inline', 'attachment'):
                disposition = 'attachment'
            # Is caching enabled or not ?
            cache = rq.get('cache', '0') == '1'
            info.writeResponse(rq.RESPONSE, obj.getDbFolder(), disposition,
                               enableCache=cache)
            # Call the "download action" if specified
            if self.downloadAction: self.downloadAction(obj.appy(), info)
# ------------------------------------------------------------------------------
